#this file was generated by __init__.py.py

import gmsh # cf. https://gitlab.onelab.info/gmsh/gmsh/-/issues/2722
import vtk

from .Constraint import *
from .Expression_MeshFunction_cpp import *
from .FOI import *
from .Kinematics import *
from .Kinematics_Inverse import *
from .Kinematics_Linearized import *
from .Loading import *
from .Material import *
from .Material_Elastic import *
from .Material_Elastic_ExponentialOgdenCiarletGeymonat import *
from .Material_Elastic_Hooke import *
from .Material_Elastic_Kirchhoff import *
from .Material_Elastic_Lung_Wbulk import *
from .Material_Elastic_Lung_Wpore import *
from .Material_Elastic_Lung_Wskel import *
from .Material_Elastic_MooneyRivlin import *
from .Material_Elastic_NeoHookean import *
from .Material_Elastic_NeoHookeanMooneyRivlin import *
from .Material_Elastic_OgdenCiarletGeymonat import *
from .Material_Elastic_OgdenCiarletGeymonatNeoHookean import *
from .Material_Elastic_OgdenCiarletGeymonatNeoHookeanMooneyRivlin import *
from .Material_Elastic_Porous import *
from .Material_Inelastic import *
from .NonlinearSolver import *
from .Operator import *
from .Operator_Constraint_MacroscopicStressComponent import *
from .Operator_HyperElasticity import *
from .Operator_HyperHydrostaticPressure import *
from .Operator_HyperIncompressibility import *
from .Operator_Inertia import *
from .Operator_LinearizedElasticity import *
from .Operator_LinearizedHydrostaticPressure import *
from .Operator_LinearizedIncompressibility import *
from .Operator_Loading_PressureBalancingGravity import *
from .Operator_Loading_SurfaceForce import *
from .Operator_Loading_SurfacePressure import *
from .Operator_Loading_SurfacePressureGradient import *
from .Operator_Loading_SurfaceTension import *
from .Operator_Loading_VolumeForce import *
from .Operator_MicroPoro_DeformedFluidVolume import *
from .Operator_MicroPoro_DeformedSolidVolume import *
from .Operator_MicroPoro_DeformedTotalVolume import *
from .Operator_Penalty_DirectionalDisplacement import *
from .Operator_Penalty_LagrangeMultiplierComponent import *
from .Operator_Penalty_MacroscopicStressComponent import *
from .Operator_Penalty_MacroscopicStretchComponent import *
from .Operator_Penalty_MacroscopicStretchSymmetry import *
from .Operator_Penalty_NormalDisplacement import *
from .Operator_Poro_Pf import *
from .Operator_Poro_Wbulk import *
from .Operator_Poro_Wpore import *
from .Operator_Poro_Wskel import *
from .Problem import *
from .Problem_Elasticity import *
from .Problem_Homogenization import *
from .Problem_Hyperelasticity import *
from .Problem_Hyperelasticity_Inverse import *
from .Problem_Hyperelasticity_MicroPoro import *
from .Problem_Hyperelasticity_Poro import *
from .Problem_Hyperelasticity_Poro_Inverse import *
from .QOI import *
from .Step import *
from .SubDomain_Periodic import *
from .SubDomain_Pinpoint import *
from .SubSol import *
from .TimeIntegrator import *
from .TimeVaryingConstant import *
from .XDMFFile import *
from .compute_error import *
from .mesh2ugrid import *
from .run_Ball_Hyperelasticity import *
from .run_Ball_Mesh import *
from .run_Disc_Hyperelasticity import *
from .run_Disc_Mesh import *
from .run_HeartSlice_Hyperelasticity import *
from .run_HeartSlice_Mesh import *
from .run_HollowBox_Homogenization import *
from .run_HollowBox_Mesh import *
from .run_HollowBox_MicroPoroHyperelasticity import *
from .run_RivlinCube_Elasticity import *
from .run_RivlinCube_Hyperelasticity import *
from .run_RivlinCube_Mesh import *
from .run_RivlinCube_PoroHyperelasticity import *
from .write_VTU_file import *



#coding=utf8

################################################################################
###                                                                          ###
### Created by Mahdi Manoochehrtayebi, 2020-2024                             ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin_mech as dmech
import myPythonLibrary as mypy

################################################################################

def run_HollowBox_Homogenization(
        dim,
        mesh = None,
        mesh_params = None,
        mat_params={},
        res_basename="run_HollowBox_Homogenization",
        write_results_to_file=1,
        verbose=0):

    ################################################################### Mesh ###

    assert ((mesh is not None) ^ (mesh_params is not None))
    if (mesh is None):
        mesh = dmech.run_HollowBox_Mesh(
            params=mesh_params)
    
    coord = mesh.coordinates()
    xmax = max(coord[:,0]); xmin = min(coord[:,0])
    ymax = max(coord[:,1]); ymin = min(coord[:,1])
    if (dim==2):    
        vol = (xmax - xmin)*(ymax - ymin)
        bbox = [xmin, xmax, ymin, ymax]
    elif (dim==3):
        zmax = max(coord[:,2]); zmin = min(coord[:,2])
        vol = (xmax - xmin)*(ymax - ymin)*(zmax - zmin)
        bbox = [xmin, xmax, ymin, ymax, zmin, zmax]

    ################################################################ Problem ###

    homogenization_problem = dmech.HomogenizationProblem(
        dim=dim,
        mesh=mesh,
        mat_params=mat_params,
        vol=vol,
        bbox=bbox)
    [mu_, lmbda_] = homogenization_problem.get_lambda_and_mu()
    kappa_ = homogenization_problem.get_kappa()

    E_ = mu_*(3*lmbda_ + 2*mu_)/(lmbda_ + mu_)
    nu_ = lmbda_/(lmbda_ + mu_)/2

    if (write_results_to_file):
        qoi_printer = mypy.DataPrinter(
            names=["E_s", "nu_s", "E_hom", "nu_hom", "kappa_hom"],
            filename=res_basename+"-qois.dat",
            limited_precision=False)
            
        qoi_printer.write_line([mat_params["E"], mat_params["nu"], E_, nu_, kappa_])
        qoi_printer.write_line([mat_params["E"], mat_params["nu"], E_, nu_, kappa_]) # MG20231124: Need to write twice for some postprocessing issue

    return mat_params["E"], mat_params["nu"], E_, nu_, kappa_



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

from curses import use_default_colors
import dolfin

import dolfin_mech as dmech

################################################################################

def run_RivlinCube_Elasticity(
        dim           : int  = 3                      ,
        incomp        : bool = 0                      ,
        multimaterial : bool = 0                      ,
        cube_params   : dict = {}                     ,
        mat_params    : dict = {}                     ,
        step_params   : dict = {}                     ,
        const_params  : dict = {}                     ,
        load_params   : dict = {}                     ,
        res_basename  : str  = "run_RivlinCube_Elasticity",
        verbose       : bool = 0                      ):

    ################################################################### Mesh ###

    if   (dim==2):
        mesh, boundaries_mf, xmin_id, xmax_id, ymin_id, ymax_id = dmech.run_RivlinCube_Mesh(dim=dim, params=cube_params)
    elif (dim==3):
        mesh, boundaries_mf, xmin_id, xmax_id, ymin_id, ymax_id, zmin_id, zmax_id = dmech.run_RivlinCube_Mesh(dim=dim, params=cube_params)

    if (multimaterial):
        mat1_sd = dolfin.CompiledSubDomain("x[0] <= x0", x0=0.5)
        mat2_sd = dolfin.CompiledSubDomain("x[0] >= x0", x0=0.5)

        mat1_id = 1
        mat2_id = 2

        domains_mf = dolfin.MeshFunction("size_t", mesh, mesh.topology().dim()) # MG20180418: size_t looks like unisgned int, but more robust wrt architecture and os
        domains_mf.set_all(0)
        mat1_sd.mark(domains_mf, mat1_id)
        mat2_sd.mark(domains_mf, mat2_id)
    else:
        domains_mf = None

    ################################################################ Problem ###

    if (incomp):
        displacement_degree = 2 # MG20211219: Incompressibility requires displacement_degree >= 2 ?!
        w_incompressibility = 1
    else:
        displacement_degree = 1
        w_incompressibility = 0

    quadrature_degree = "default"
    # quadrature_degree = "full"

    if (multimaterial):
        elastic_behavior = None
        if (incomp):
            mat1_mod = "H_dev"
            mat2_mod = "H_dev"
        else:
            mat1_mod = "H"
            mat2_mod = "H"
        mat1_params = {
            "E":1.,
            "nu":0.5*(incomp)+0.3*(1-incomp)}

        mat2_params = {
            "E":10.,
            "nu":0.5*(incomp)+0.3*(1-incomp)}
        elastic_behaviors=[
                {"subdomain_id":mat1_id, "model":mat1_mod, "parameters":mat1_params, "suffix":"1"},
                {"subdomain_id":mat2_id, "model":mat2_mod, "parameters":mat2_params, "suffix":"2"}]
    else:
        elastic_behavior = mat_params
        elastic_behaviors = None

    problem = dmech.ElasticityProblem(
        mesh=mesh,
        domains_mf=domains_mf,
        define_facet_normals=1,
        boundaries_mf=boundaries_mf,
        displacement_degree=displacement_degree,
        quadrature_degree=quadrature_degree,
        w_incompressibility=w_incompressibility,
        elastic_behavior=elastic_behavior,
        elastic_behaviors=elastic_behaviors)

    ########################################## Boundary conditions & Loading ###

    const_type = const_params.get("type", "sym")

    if (const_type in ("symx", "sym")):
        problem.add_constraint(V=problem.get_displacement_function_space().sub(0), sub_domains=boundaries_mf, sub_domain_id=xmin_id, val=0.)
    if (const_type in ("symy", "sym")) and (dim >= 2):
        problem.add_constraint(V=problem.get_displacement_function_space().sub(1), sub_domains=boundaries_mf, sub_domain_id=ymin_id, val=0.)
    if (const_type in ("symz", "sym")) and (dim >= 3):
        problem.add_constraint(V=problem.get_displacement_function_space().sub(2), sub_domains=boundaries_mf, sub_domain_id=zmin_id, val=0.)
    if (const_type in ("blox")):
        problem.add_constraint(V=problem.get_displacement_function_space(), sub_domains=boundaries_mf, sub_domain_id=xmin_id, val=[0.]*dim)
    if (const_type in ("bloy")):
        problem.add_constraint(V=problem.get_displacement_function_space(), sub_domains=boundaries_mf, sub_domain_id=ymin_id, val=[0.]*dim)
    if (const_type in ("bloz")):
        problem.add_constraint(V=problem.get_displacement_function_space(), sub_domains=boundaries_mf, sub_domain_id=zmin_id, val=[0.]*dim)

    load_type = load_params.get("type", "disp")

    Deltat = step_params.get("Deltat", 1.)
    dt_ini = step_params.get("dt_ini", 1.)
    dt_min = step_params.get("dt_min", 1.)

    k_step = problem.add_step(
        Deltat=Deltat,
        dt_ini=dt_ini,
        dt_min=dt_min)

    if (load_type == "disp"):
        u = load_params.get("u", 1.)
        problem.add_constraint(
            V=problem.get_displacement_function_space().sub(0),
            sub_domains=boundaries_mf,
            sub_domain_id=xmax_id,
            val_ini=0., val_fin=u,
            k_step=k_step)
    elif (load_type == "volu"):
        f = load_params.get("f", 1.)
        problem.add_volume_force0_loading_operator(
            measure=problem.dV,
            F_ini=[0.]*dim, F_fin=[f]+[0.]*(dim-1),
            k_step=k_step)
    elif (load_type == "surf"):
        f = load_params.get("f", 1.)
        problem.add_surface_force0_loading_operator(
            measure=problem.dS(xmax_id),
            F_ini=[0.]*dim, F_fin=[f]+[0.]*(dim-1),
            k_step=k_step)
    elif (load_type == "pres"):
        p = load_params.get("p", -1.)
        problem.add_surface_pressure0_loading_operator(
            measure=problem.dS(xmax_id),
            P_ini=0, P_fin=p,
            k_step=k_step)
    elif (load_type == "pgra"):
        X0 = load_params.get("X0", [0.5]*dim)
        N0 = load_params.get("N0", [1.]+[0.]*(dim-1))
        P0 = load_params.get("P0", -1.0)
        DP = load_params.get("DP", -0.5)
        problem.add_surface_pressure_gradient0_loading_operator(
            measure=problem.dS(),
            X0_val=X0,
            N0_val=N0,
            P0_ini=0., P0_fin=P0,
            DP_ini=0., DP_fin=DP,
            k_step=k_step)
    elif (load_type == "tens"):
        gamma = load_params.get("gamma", 0.01)
        problem.add_surface_tension0_loading_operator(
            measure=problem.dS,
            gamma_ini=0.0, gamma_fin=gamma,
            k_step=k_step)

    ################################################# Quantities of Interest ###

    problem.add_global_strain_qois()
    problem.add_global_stress_qois()
    if (incomp): problem.add_global_pressure_qoi()

    ################################################################# Solver ###

    solver = dmech.NonlinearSolver(
        problem=problem,
        parameters={
            "sol_tol":[1e-6]*len(problem.subsols),
            "n_iter_max":32},
        relax_type="constant",
        write_iter=0)

    integrator = dmech.TimeIntegrator(
        problem=problem,
        solver=solver,
        parameters={
            "n_iter_for_accel":4,
            "n_iter_for_decel":16,
            "accel_coeff":2,
            "decel_coeff":2},
        print_out=res_basename*verbose,
        print_sta=res_basename*verbose,
        write_qois=res_basename+"-qois",
        write_qois_limited_precision=1,
        write_sol=res_basename*verbose)

    success = integrator.integrate()
    assert (success),\
        "Integration failed. Aborting."

    integrator.close()



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import gmsh
import meshio

################################################################################

def run_Disc_Mesh(
        params : dict = {}):

    X0 = params.get("X0", 0.5)
    Y0 = params.get("Y0", 0.5)
    R  = params.get("R" , 0.3)
    l  = params.get("l" , 0.1)

    mesh_filebasename = params.get("mesh_filebasename", "mesh")

    ################################################################### Mesh ###
    
    gmsh.initialize()
    gmsh.clear()
    factory = gmsh.model.geo

    p0 = factory.addPoint(x=X0  , y=Y0  , z=0, meshSize=l)
    p1 = factory.addPoint(x=X0+R, y=Y0  , z=0, meshSize=l)
    p2 = factory.addPoint(x=X0  , y=Y0+R, z=0, meshSize=l)
    p3 = factory.addPoint(x=X0-R, y=Y0  , z=0, meshSize=l)
    p4 = factory.addPoint(x=X0  , y=Y0-R, z=0, meshSize=l)

    l1 = factory.addCircleArc(p1, p0, p2)
    l2 = factory.addCircleArc(p2, p0, p3)
    l3 = factory.addCircleArc(p3, p0, p4)
    l4 = factory.addCircleArc(p4, p0, p1)

    cl = factory.addCurveLoop([l1, l2, l3, l4])

    s = factory.addPlaneSurface([cl])

    factory.synchronize()

    ps = gmsh.model.addPhysicalGroup(dim=2, tags=[s])

    mesh_gmsh = gmsh.model.mesh

    mesh_gmsh.generate(dim=2)

    gmsh.write(mesh_filebasename+".vtk")

    gmsh.finalize()

    mesh_meshio = meshio.read(mesh_filebasename+".vtk")

    mesh_meshio.points = mesh_meshio.points[:, :2]

    meshio.write(mesh_filebasename+".xdmf", mesh_meshio)

    mesh = dolfin.Mesh()
    dolfin.XDMFFile(mesh_filebasename+".xdmf").read(mesh)

    dolfin.File(mesh_filebasename+".xml") << mesh

    ############################################################# Boundaries ###

    boundaries_mf = dolfin.MeshFunction(
        value_type="size_t",
        mesh=mesh,
        dim=1)

    boundaries_mf.set_all(0)

    S_sd = dolfin.AutoSubDomain(
        lambda x, on_boundary:
            on_boundary and\
            dolfin.near(
                (x[0]-X0)**2 + (x[1]-Y0)**2,
                R**2,
                eps=1e-3))

    S_id = 1; S_sd.mark(boundaries_mf, S_id)

    # dolfin.XDMFFile(mesh_filebasename+"-boundaries.xdmf").write(boundaries_mf)

    ################################################################# Points ###

    points_mf = dolfin.MeshFunction(
        value_type="size_t",
        mesh=mesh,
        dim=0)

    points_mf.set_all(0)

    x1 = [X0+R, Y0]
    x1_sd = dolfin.AutoSubDomain(
        lambda x, on_boundary:
            dolfin.near(x[0], x1[0], eps=1e-3)
        and dolfin.near(x[1], x1[1], eps=1e-3))
    x2 = [X0, Y0+R]
    x2_sd = dolfin.AutoSubDomain(
        lambda x, on_boundary:
            dolfin.near(x[0], x2[0], eps=1e-3)
        and dolfin.near(x[1], x2[1], eps=1e-3))
    x3 = [X0-R, Y0]
    x3_sd = dolfin.AutoSubDomain(
        lambda x, on_boundary:
            dolfin.near(x[0], x3[0], eps=1e-3)
        and dolfin.near(x[1], x3[1], eps=1e-3))
    x4 = [X0, Y0-R]
    x4_sd = dolfin.AutoSubDomain(
        lambda x, on_boundary:
            dolfin.near(x[0], x4[0], eps=1e-3)
        and dolfin.near(x[1], x4[1], eps=1e-3))

    x1_id = 1; x1_sd.mark(points_mf, x1_id)
    x2_id = 2; x2_sd.mark(points_mf, x2_id)
    x3_id = 3; x3_sd.mark(points_mf, x3_id)
    x4_id = 4; x4_sd.mark(points_mf, x4_id)

    # dolfin.XDMFFile(mesh_filebasename+"-points.xdmf").write(points_mf)

    return mesh, boundaries_mf, S_id, points_mf, x1_sd, x2_sd, x3_sd, x4_sd



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import os
import shutil

import myVTKPythonLibrary as myvtk

import dolfin_mech as dmech

################################################################################

def write_VTU_file(
        filebasename,
        function,
        time=None,
        zfill=3,
        preserve_connectivity=False):

    if (preserve_connectivity):
        ugrid = dmech.mesh2ugrid(function.function_space().mesh())
        dmech.add_function_to_ugrid(
            function=function,
            ugrid=ugrid)
        myvtk.writeUGrid(
            ugrid=ugrid,
            filename=filebasename+("_"+str(time).zfill(zfill) if (time is not None) else "")+".vtu")

    else:
        file_pvd = dolfin.File(filebasename+"__.pvd")
        file_pvd << (function, float(time) if (time is not None) else 0.)
        os.remove(
            filebasename+"__.pvd")
        shutil.move(
            filebasename+"__"+"".zfill(6)+".vtu",
            filebasename+("_"+str(time).zfill(zfill) if (time is not None) else "")+".vtu")



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech

################################################################################

class Step():



    def __init__(self,
            t_ini=0.,
            t_fin=1.,
            dt_ini=None,
            dt_min=None,
            dt_max=None,
            operators=None,    # MG20180508: Do not use list as default value because it is static
            constraints=None): # MG20180508: Do not use list as default value because it is static

        self.t_ini = t_ini
        self.t_fin = t_fin

        self.dt_ini = dt_ini if (dt_ini is not None) else self.t_fin - self.t_ini
        self.dt_min = dt_min if (dt_min is not None) else self.dt_ini
        self.dt_max = dt_max if (dt_max is not None) else self.dt_ini

        self.operators   = operators   if (operators   is not None) else []
        self.constraints = constraints if (constraints is not None) else []



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class SurfacePressureGradientLoadingOperator(Operator):

    def __init__(self,
            X,
            U,
            U_test,
            kinematics,
            N,
            measure,
            X0_val=None, X0_ini=None, X0_fin=None,
            N0_val=None, N0_ini=None, N0_fin=None,
            P0_val=None, P0_ini=None, P0_fin=None,
            DP_val=None, DP_ini=None, DP_fin=None):

        self.measure = measure

        self.tv_X0 = dmech.TimeVaryingConstant(
            val=X0_val, val_ini=X0_ini, val_fin=X0_fin)
        X0 = self.tv_X0.val
        self.tv_N0 = dmech.TimeVaryingConstant(
            val=N0_val, val_ini=N0_ini, val_fin=N0_fin)
        N0 = self.tv_N0.val
        self.tv_P0 = dmech.TimeVaryingConstant(
            val=P0_val, val_ini=P0_ini, val_fin=P0_fin)
        P0 = self.tv_P0.val
        self.tv_DP = dmech.TimeVaryingConstant(
            val=DP_val, val_ini=DP_ini, val_fin=DP_fin)
        DP = self.tv_DP.val

        x = X + U
        P = P0 + DP * dolfin.inner(x - X0, N0)

        T = dolfin.dot(-P * N, dolfin.inv(kinematics.F))
        self.res_form = - dolfin.inner(T, U_test) * kinematics.J * self.measure



    def set_value_at_t_step(self,
            t_step):

        self.tv_X0.set_value_at_t_step(t_step)
        self.tv_N0.set_value_at_t_step(t_step)
        self.tv_P0.set_value_at_t_step(t_step)
        self.tv_DP.set_value_at_t_step(t_step)

################################################################################

class SurfacePressureGradient0LoadingOperator(Operator):

    def __init__(self,
            x,
            U_test,
            N,
            measure,
            X0_val=None, X0_ini=None, X0_fin=None,
            N0_val=None, N0_ini=None, N0_fin=None,
            P0_val=None, P0_ini=None, P0_fin=None,
            DP_val=None, DP_ini=None, DP_fin=None):

        self.measure = measure

        self.tv_X0 = dmech.TimeVaryingConstant(
            val=X0_val, val_ini=X0_ini, val_fin=X0_fin)
        X0 = self.tv_X0.val
        self.tv_N0 = dmech.TimeVaryingConstant(
            val=N0_val, val_ini=N0_ini, val_fin=N0_fin)
        N0 = self.tv_N0.val
        self.tv_P0 = dmech.TimeVaryingConstant(
            val=P0_val, val_ini=P0_ini, val_fin=P0_fin)
        P0 = self.tv_P0.val
        self.tv_DP = dmech.TimeVaryingConstant(
            val=DP_val, val_ini=DP_ini, val_fin=DP_fin)
        DP = self.tv_DP.val

        P = P0 + DP * dolfin.inner(x - X0, N0)

        self.res_form = - dolfin.inner(-P * N, U_test) * self.measure



    def set_value_at_t_step(self,
            t_step):

        self.tv_X0.set_value_at_t_step(t_step)
        self.tv_N0.set_value_at_t_step(t_step)
        self.tv_P0.set_value_at_t_step(t_step)
        self.tv_DP.set_value_at_t_step(t_step)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import numpy

import dolfin_mech as dmech

################################################################################

class TimeVaryingConstant():



    def __init__(self,
            val=None, val_ini=None, val_fin=None):

        if  (val     is not None)\
        and (val_ini is     None)\
        and (val_fin is     None):
            val_ini = val
            val_fin = val
        elif (val     is     None)\
         and (val_ini is not None)\
         and (val_fin is not None):
            pass
        else:
            assert (0), "Must provide val or val_ini & val_fin. Aborting."

        assert (type(val_ini) in (int, float, list, numpy.ndarray)),\
             "type(val_ini) (="+str(type(val_ini))+") should be int, float, list or numpy.ndarray. Aborting."
        if (type(val_ini) in (int, float)):
            assert (type(val_fin) in (int, float)),\
             "type(val_fin) (="+str(type(val_fin))+") should be int or float. Aborting."
            self.val_ini = numpy.array([val_ini], dtype=float)
            self.val_fin = numpy.array([val_fin], dtype=float)
            self.val_cur = numpy.array([val_ini], dtype=float)
            self.val_old = numpy.array([val_ini], dtype=float)
            self.set_value = self.set_value_sca
        elif (type(val_ini) in (list, numpy.ndarray)):
            assert (type(val_fin) in (list, numpy.ndarray)),\
             "type(val_fin) (="+str(type(val_fin))+") should be list or numpy.ndarray. Aborting."
            self.val_ini = numpy.array(val_ini, dtype=float)
            self.val_fin = numpy.array(val_fin, dtype=float)
            self.val_cur = numpy.array(val_ini, dtype=float)
            self.val_old = numpy.array(val_ini, dtype=float)
            self.set_value = self.set_value_vec
        self.val = dolfin.Constant(val_ini)
        # print("ini", self.val_ini)
        # print("fin", self.val_fin)



    def set_value_sca(self,
            val):

        if   (type(val) in (int, float)):
            self.val.assign(dolfin.Constant(val))
        elif (type(val) in (list, numpy.ndarray)):
            self.val.assign(dolfin.Constant(val[0]))



    def set_value_vec(self,
            val):

        # print(val)
        self.val.assign(dolfin.Constant(val))



    def set_value_at_t_step(self,
            t_step):

        # print("set_value_at_t_step")
        # print("t_step", t_step)
        # print("ini", self.val_ini)
        # print("fin", self.val_fin)
        # print("cur", self.val_ini * (1. - t_step) + self.val_fin * t_step)
        self.set_value(self.val_ini * (1. - t_step) + self.val_fin * t_step)



    def set_dvalue_at_t_step(self,
            t_step):

        # print("set_dvalue_at_t_step")
        self.val_old[:] = self.val_cur[:]
        # print("old", self.val_old)
        # print("t_step", t_step)
        # print("ini", self.val_ini)
        # print("fin", self.val_fin)
        self.val_cur[:] = self.val_ini * (1. - t_step) + self.val_fin * t_step
        # print("cur", self.val_cur)
        # print("dvalue", self.val_cur - self.val_old)
        # print(self.val.str(1))
        self.set_value(self.val_cur - self.val_old)
        # print(self.val.str(1))



    def restore_old_value(self):

        self.val_cur[:] = self.val_old[:]



    def homogenize(self):

        # print("homogenize")
        # print(self.val.str(1))
        self.set_value(0*self.val_ini)
        # print(self.val.str(1))



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import math
import numpy

import dolfin_mech as dmech

################################################################################

def run_Ball_Hyperelasticity(
        incomp=0,
        mesh_params={},
        mat_params={},
        step_params={},
        load_params={},
        res_basename="run_Ball_Hyperelasticity",
        write_vtus_with_preserved_connectivity=False,
        verbose=0):

    ################################################################### Mesh ###

    X0 = mesh_params.get("X0", 0.5)
    Y0 = mesh_params.get("Y0", 0.5)
    Z0 = mesh_params.get("Z0", 0.5)
    R  = mesh_params.get("R" , 0.3)

    mesh, boundaries_mf, S_id = dmech.run_Ball_Mesh(
        params=mesh_params)

    ################################################################ Problem ###

    if (incomp):
        displacement_degree = 2 # MG20211219: Incompressibility requires displacement_degree >= 2 ?!
    else:
        displacement_degree = 1

    problem = dmech.HyperelasticityProblem(
        mesh=mesh,
        define_facet_normals=1,
        boundaries_mf=boundaries_mf,
        displacement_degree=displacement_degree,
        quadrature_degree="default",
        w_incompressibility=incomp,
        elastic_behavior=mat_params)

    ########################################## Boundary conditions & Loading ###

    Deltat = step_params.get("Deltat", 1.)
    dt_ini = step_params.get("dt_ini", 1.)
    dt_min = step_params.get("dt_min", 1.)

    k_step = problem.add_step(
        Deltat=Deltat,
        dt_ini=dt_ini,
        dt_min=dt_min)

    load_type = load_params.get("type", "disp")

    if (load_type == "disp"): # MG20220813: It would be possible to impose the spatially varying displacement directly through an expression, but this would need to be implemented within Constraint, e.g. with a TimeVaryingExpression.
        surface_nodes_coords = [node_coords for node_coords in mesh.coordinates() if dolfin.near((node_coords[0]-X0)**2 + (node_coords[1]-Y0)**2 + (node_coords[2]-Z0)**2, R**2, eps=1e-3)]
        dR = load_params.get("dR", +0.1)
        for X in surface_nodes_coords:
            eR = numpy.subtract(X, [X0, Y0, Z0])
            eR /= numpy.linalg.norm(eR)
            U = dR * eR
            # X_sd = dolfin.AutoSubDomain(lambda x, on_boundary: dolfin.near(x[0], X[0], eps=1e-3) and dolfin.near(x[1], X[1], eps=1e-3)) # MG20220813: OMG this behaves so weird!
            X_sd = dolfin.CompiledSubDomain("near(x[0], x0) && near(x[1], y0) && near(x[2], z0)", x0=X[0], y0=X[1], z0=X[2])
            problem.add_constraint(
                V=problem.get_displacement_function_space(),
                sub_domain=X_sd,
                val_ini=[0.,0.,0.], val_fin=U,
                k_step=k_step,
                method="pointwise")

    ################################################# Quantities of Interest ###

    problem.add_point_displacement_qoi(
        name="U",
        coordinates=[X0+R, Y0, Z0],
        component=0)

    ################################################################# Solver ###

    solver = dmech.NonlinearSolver(
        problem=problem,
        parameters={
            "sol_tol":[1e-6]*len(problem.subsols),
            "n_iter_max":32},
        relax_type="constant",
        write_iter=0)

    integrator = dmech.TimeIntegrator(
        problem=problem,
        solver=solver,
        parameters={
            "n_iter_for_accel":4,
            "n_iter_for_decel":16,
            "accel_coeff":2,
            "decel_coeff":2},
        print_out=res_basename*verbose,
        print_sta=res_basename*verbose,
        write_qois=res_basename+"-qois",
        write_qois_limited_precision=1,
        write_sol=res_basename*verbose,
        write_vtus=res_basename*verbose,
        write_vtus_with_preserved_connectivity=write_vtus_with_preserved_connectivity)

    success = integrator.integrate()
    assert (success),\
        "Integration failed. Aborting."

    integrator.close()



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import sys

import myPythonLibrary as mypy

import dolfin_mech as dmech

################################################################################

class TimeIntegrator():

    def __init__(self,
            problem,
            solver,
            parameters,
            print_out=True,
            print_sta=True,
            write_qois=True,
            write_qois_limited_precision=False,
            write_sol=True,
            write_vtus=False,
            write_vtus_with_preserved_connectivity=False,
            write_xmls=False):

        self.problem = problem

        self.solver = solver

        self.n_iter_for_accel = parameters.get("n_iter_for_accel",  4)
        self.n_iter_for_decel = parameters.get("n_iter_for_decel", 16)
        self.accel_coeff      = parameters.get("accel_coeff"     ,  2)
        self.decel_coeff      = parameters.get("decel_coeff"     ,  2)

        if (type(print_out) is str):
            if (print_out=="stdout"):
                self.printer_filename = None
            elif (print_out=="argv"):
                self.printer_filename = sys.argv[0][:-3]+".out"
            else:
                self.printer_filename = print_out+".out"
        else:
            self.printer_filename = None
        self.printer = mypy.Printer(
            filename=self.printer_filename,
            silent=not(print_out))
        self.solver.printer = self.printer

        if (type(print_sta) is str):
            if (print_sta=="stdout"):
                self.table_printer_filename = None
            elif (print_sta=="argv"):
                self.table_printer_filename = sys.argv[0][:-3]+".sta"
            else:
                self.table_printer_filename = print_sta+".sta"
        else:
            self.table_printer_filename = sys.argv[0][:-3]+".sta"
        self.table_printer = mypy.TablePrinter(
            titles=["k_step", "k_t", "dt", "t", "t_step", "n_iter", "success"],
            filename=self.table_printer_filename,
            silent=not(print_sta))

        self.write_qois = bool(write_qois) and (len(self.problem.qois)>0)
        if (self.write_qois):
            self.write_qois_filebasename = write_qois if (type(write_qois) is str) else sys.argv[0][:-3]+"-qois"

            self.qoi_printer = mypy.DataPrinter(
                names=["t"]+[qoi.name for qoi in self.problem.qois],
                filename=self.write_qois_filebasename+".dat",
                limited_precision=write_qois_limited_precision)

            self.problem.update_qois(dt=1)
            self.qoi_printer.write_line([0.]+[qoi.value for qoi in self.problem.qois])

        self.problem.update_fois()
        self.write_sol = bool(write_sol)
        if (self.write_sol):
            self.write_sol_filebasename = write_sol if (type(write_sol) is str) else sys.argv[0][:-3]+"-sol"

            self.functions_to_write  = []
            self.functions_to_write += self.problem.get_subsols_func_lst()
            self.functions_to_write += self.problem.get_subsols_func_old_lst()
            self.functions_to_write += self.problem.get_fois_func_lst()

            self.xdmf_file_sol = dmech.XDMFFile(
                filename=self.write_sol_filebasename+".xdmf",
                functions=self.functions_to_write)
            self.xdmf_file_sol.write(0.)

            self.write_vtus                             = bool(write_vtus)
            self.write_vtus_with_preserved_connectivity = bool(write_vtus_with_preserved_connectivity)
            if (self.write_vtus):
                dmech.write_VTU_file(
                    filebasename=self.write_sol_filebasename,
                    function=self.problem.get_displacement_subsol().subfunc,
                    time=0,
                    preserve_connectivity=self.write_vtus_with_preserved_connectivity)

            self.write_xmls = bool(write_xmls)
            if (self.write_xmls):
                dolfin.File(self.write_sol_filebasename+"_"+str(0).zfill(3)+".xml") << self.problem.get_displacement_subsol().subfunc



    def close(self):

        self.printer.close()
        self.table_printer.close()

        if (self.write_qois):
            self.qoi_printer.close()

        if (self.write_sol):
            self.xdmf_file_sol.close()



    def integrate(self):

        k_t_tot = 0
        n_iter_tot = 0
        self.printer.inc()
        for k_step in range(1,len(self.problem.steps)+1):
            self.printer.print_var("k_step",k_step,-1)

            self.step = self.problem.steps[k_step-1]

            t = self.step.t_ini
            dt = self.step.dt_ini

            self.problem.set_variational_formulation(
                k_step=k_step-1)

            self.solver.constraints  = []
            self.solver.constraints += self.problem.constraints
            self.solver.constraints += self.step.constraints

            k_t = 0
            self.printer.inc()
            while (True):
                k_t += 1
                k_t_tot += 1
                self.printer.print_var("k_t",k_t,-1)

                if (t+dt > self.step.t_fin):
                    dt = self.step.t_fin - t
                self.printer.print_var("dt",dt)

                # self.problem.set_variational_formulation(
                #     k_step=k_step-1,
                #     dt=dt)

                t += dt
                self.printer.print_var("t",t)

                t_step = (t - self.step.t_ini)/(self.step.t_fin - self.step.t_ini)
                self.printer.print_var("t_step",t_step)

                for operator in self.step.operators:
                    operator.set_value_at_t_step(t_step)
                    operator.set_dt(dt)

                for constraint in self.step.constraints:
                    constraint.set_value_at_t_step(t_step)

                for inelastic_behavior in self.problem.inelastic_behaviors_internal:
                    inelastic_behavior.update_internal_variables_at_t(t)

                self.problem.sol_old_func.vector()[:] = self.problem.sol_func.vector()[:]
                if (len(self.problem.subsols) > 1):
                    dolfin.assign(
                        self.problem.get_subsols_func_old_lst(),
                        self.problem.sol_old_func)
                solver_success, n_iter = self.solver.solve(k_step, k_t, dt, t)

                self.table_printer.write_line([k_step, k_t, dt, t, t_step, n_iter, solver_success])

                if (solver_success):
                    n_iter_tot += n_iter

                    self.problem.update_fois()
                    if (self.write_sol):
                        self.xdmf_file_sol.write(t)

                        if (self.write_vtus):
                            dmech.write_VTU_file(
                                filebasename=self.write_sol_filebasename,
                                function=self.problem.get_displacement_subsol().subfunc,
                                time=k_t_tot,
                                preserve_connectivity=self.write_vtus_with_preserved_connectivity)

                        if (self.write_xmls):
                            dolfin.File(self.write_sol_filebasename+"_"+str(k_t_tot).zfill(3)+".xml") << self.problem.get_displacement_subsol().subfunc

                    if (self.write_qois):
                        self.problem.update_qois(dt)
                        self.qoi_printer.write_line([t]+[qoi.value for qoi in self.problem.qois])

                    if dolfin.near(t, self.step.t_fin, eps=1e-9):
                        self.success = True
                        break
                    else:
                        if (n_iter <= self.n_iter_for_accel):
                            dt *= self.accel_coeff
                            if (dt > self.step.dt_max):
                                dt = self.step.dt_max
                        elif (n_iter >= self.n_iter_for_decel):
                            dt /= self.decel_coeff
                            if (dt < self.step.dt_min):
                                dt = self.step.dt_min
                else:
                    self.problem.sol_func.vector()[:] = self.problem.sol_old_func.vector()[:]
                    if (len(self.problem.subsols) > 1):
                        dolfin.assign(
                            self.problem.get_subsols_func_lst(),
                            self.problem.sol_func)

                    for inelastic_behavior in self.problem.inelastic_behaviors_internal:
                        inelastic_behavior.restore_old_value()

                    for constraint in self.step.constraints:
                        constraint.restore_old_value()

                    k_t -= 1
                    k_t_tot -= 1
                    t -= dt

                    dt /= self.decel_coeff
                    if (dt < self.step.dt_min):
                        self.printer.print_str("Warning! Time integrator failed to move forward!")
                        self.success = False
                        break

            self.printer.dec()

            if not (self.success):
                break

        self.printer.dec()

        return self.success



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import numpy

import dolfin_mech as dmech

################################################################################

class SubSol():



    def __init__(self,
            name,
            fe,
            init_val=None,
            init_fun=None):

        self.name = name
        self.fe = fe

        if   (init_val is     None) and (init_fun is     None):
            self.init_val = numpy.zeros(fe.value_shape())
            self.init = self.init_with_val
        elif (init_val is not None) and (init_fun is     None):
            assert (numpy.shape(init_val) == self.fe.value_shape())
            self.init_val = numpy.asarray(init_val)
            self.init = self.init_with_val
        elif (init_val is     None) and (init_fun is not None):
            self.init_fun = init_fun
            self.init = self.init_with_field
        else:
            assert (0), "Can only provide init_val or init_fun. Aborting."



    def init_with_val(self):
        init_val_str = self.init_val.astype(str).tolist()
        # print(self.func.vector().get_local())
        self.func.interpolate(dolfin.Expression(
            init_val_str,
            element=self.fe))
        # print(self.func.vector().get_local())
        self.func_old.interpolate(dolfin.Expression(
            init_val_str,
            element=self.fe))



    def init_with_field(self):
        self.func.vector()[:]     = self.init_fun.vector().get_local()
        self.func_old.vector()[:] = self.init_fun.vector().get_local()



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class WskelPoroOperator(Operator):

    def __init__(self,
            kinematics,
            U_test,
            Phis0,
            material_parameters,
            material_scaling,
            measure):

        self.kinematics = kinematics
        self.solid_material = dmech.WskelLungElasticMaterial(
            kinematics=kinematics,
            parameters=material_parameters)
        self.material = dmech.PorousElasticMaterial(
            solid_material=self.solid_material,
            scaling=material_scaling,
            Phis0=Phis0)
        self.measure = measure

        dE_test = dolfin.derivative(
            self.kinematics.E, self.kinematics.U, U_test)
        self.res_form = dolfin.inner(self.material.Sigma, dE_test) * self.measure

################################################################################

class InverseWskelPoroOperator(Operator):

    def __init__(self,
            kinematics,
            u_test,
            phis0,
            material_parameters,
            material_scaling,
            measure):

        self.kinematics = kinematics
        self.solid_material = dmech.WskelLungElasticMaterial(
            kinematics=kinematics,
            parameters=material_parameters)
        self.material = dmech.PorousElasticMaterial(
            solid_material=self.solid_material,
            scaling=material_scaling,
            Phis0=self.kinematics.J * phis0)
        self.measure = measure

        epsilon_test = dolfin.sym(dolfin.grad(u_test))
        self.res_form = dolfin.inner(self.material.sigma, epsilon_test) * self.measure



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech

################################################################################

class Loading():



    def __init__(self,
            measure,
            val=None,
            val_ini=None, val_fin=None,
            xyz_ini=None,
            N=None):

        self.measure = measure

        if (val is not None) and (val_ini is None) and (val_fin is None):
            self.tv_val = dmech.TimeVaryingConstant(
                val_ini=val, val_fin=val)
        elif (val is None) and (val_ini is not None) and (val_fin is not None):
            self.tv_val = dmech.TimeVaryingConstant(
                val_ini=val_ini, val_fin=val_fin)
        self.val = self.tv_val.val

        if (N is not None):
            self.N = dolfin.Constant(N)

        if (xyz_ini is not None):
            self.xyz_ini = dolfin.Constant(xyz_ini)



    def set_value_at_t_step(self,
            t_step):

        self.tv_val.set_value_at_t_step(t_step)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Felipe Álvarez-Barrientos, 2019-2021                                 ###
###                                                                          ###
### Pontificia Universidad Católica, Santiago, Chile                         ###
###                                                                          ###
###                                                                          ###
### And Mahdi Manoochehrtayebi, 2020-2024                                    ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import numpy

################################################################################

class PinpointSubDomain(dolfin.SubDomain):



    def __init__(self, coords, tol=None):

        self.coords = numpy.asarray(coords)
        self.tol = tol if tol is not None else 1e-3

        dolfin.SubDomain.__init__(self)



    def move(self, coords):

        self.coords[:] = coords



    def inside(self, x, on_boundary):

        return (numpy.linalg.norm(x - self.coords) < self.tol)



    def check_inside(self, mesh):

        x_lst = []
        for x in mesh.coordinates():
            if self.inside(x, True):
                x_lst.append(x)
        return x_lst



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech

################################################################################

class Material():



    def get_lambda_from_parameters(self,
            parameters):

        if ("lambda" in parameters):
            lmbda = parameters["lambda"]
        elif ("E" in parameters) and ("nu" in parameters):
            E  = parameters["E"]
            nu = parameters["nu"]
            if parameters.get("PS", False):
                lmbda = E*nu/(1+nu)/(1-  nu)
            else:
                lmbda = E*nu/(1+nu)/(1-2*nu)
        else:
            assert (0),\
                "No parameter found: \"+str(parameters)+\". Must provide lambda or E & nu. Aborting."
        return dolfin.Constant(lmbda)



    def get_mu_from_parameters(self,
            parameters):

        if ("mu" in parameters):
            mu = parameters["mu"]
        elif ("E" in parameters) and ("nu" in parameters):
            E  = parameters["E"]
            nu = parameters["nu"]
            mu = E/2/(1+nu)
        else:
            assert (0),\
                "No parameter found: \"+str(parameters)+\". Must provide mu or E & nu. Aborting."
        return dolfin.Constant(mu)



    def get_lambda_and_mu_from_parameters(self,
            parameters):

        lmbda = self.get_lambda_from_parameters(parameters)
        mu    = self.get_mu_from_parameters(parameters)

        return lmbda, mu



    def get_K_from_parameters(self,
            parameters):

        if ("K" in parameters):
            K = parameters["K"]
        else:
            lmbda, mu = self.get_lambda_and_mu_from_parameters(parameters)
            K = (3*lmbda+2*mu)/3
        return dolfin.Constant(K)



    def get_G_from_parameters(self,
            parameters):

        if ("G" in parameters):
            G = parameters["G"]
        else:
            mu = self.get_mu_from_parameters(parameters)
            G = mu
        return dolfin.Constant(G)


    
    def get_C0_from_parameters(self,
            parameters,
            decoup=False):

        if ("C0" in parameters):
            C0 = parameters["C0"]
        elif ("c0" in parameters):
            C0 = parameters["c0"]
        else:
            if (decoup):
                K = self.get_K_from_parameters(parameters)
                C0 = K/4
            else:
                lmbda = self.get_lambda_from_parameters(parameters)
                C0 = lmbda/4
        return dolfin.Constant(C0)


    
    def get_C1_from_parameters(self,
            parameters):

        if ("C1" in parameters):
            C1 = parameters["C1"]
        elif ("c1" in parameters):
            C1 = parameters["c1"]
        elif ("mu" in parameters):
            mu = parameters["mu"]
            C1 = mu/2
        elif ("E" in parameters) and ("nu" in parameters):
            E  = parameters["E"]
            nu = parameters["nu"]
            mu = E/2/(1+nu)
            C1 = mu/2
        else:
            assert (0),\
                "No parameter found ("+str(parameters)+"). Must provide C1 or c1 or mu or E & nu. Aborting."
        return dolfin.Constant(C1)



    def get_C2_from_parameters(self,
            parameters):

        if ("C2" in parameters):
            C2 = parameters["C2"]
        elif ("c2" in parameters):
            C2 = parameters["c2"]
        elif ("mu" in parameters):
            mu = parameters["mu"]
            C2 = mu/2
        elif ("E" in parameters) and ("nu" in parameters):
            E  = parameters["E"]
            nu = parameters["nu"]
            mu = E/2/(1+nu)
            C2 = mu/2
        else:
            assert (0),\
                "No parameter found ("+str(parameters)+"). Must provide C2 or c2 or mu or E & nu. Aborting."
        return dolfin.Constant(C2)



    def get_C1_and_C2_from_parameters(self,
            parameters):

        if ("C1" in parameters) and ("C2" in parameters):
            C1 = parameters["C1"]
            C2 = parameters["C2"]
        elif ("c1" in parameters) and ("c2" in parameters):
            C1 = parameters["c1"]
            C2 = parameters["c2"]
        elif ("mu" in parameters):
            mu = parameters["mu"]
            C1 = mu/4
            C2 = mu/4
        elif ("E" in parameters) and ("nu" in parameters):
            E  = parameters["E"]
            nu = parameters["nu"]
            mu = E/2/(1+nu)
            C1 = mu/4
            C2 = mu/4
        else:
            assert (0),\
                "No parameter found ("+str(parameters)+"). Must provide C1 & C2 or c1 & c2 or mu or E & nu. Aborting."
        return dolfin.Constant(C1), dolfin.Constant(C2)



################################################################################



    # def get_C_from_U_or_C(self,
    #         U=None,
    #         C=None):

    #     if (C is not None):
    #         assert (C.ufl_shape[0] == C.ufl_shape[1])
    #     elif (U is not None):
    #         dim = U.ufl_shape[0]
    #         I = dolfin.Identity(dim)
    #         F = I + dolfin.grad(U)
    #         # JF = dolfin.det(F) # MG20211220: Otherwise cannot derive Psi wrt C
    #         C = F.T * F
    #     else:
    #         assert (0),\
    #             "Must provide U or C. Aborting."
    #     return dolfin.variable(C)



    # def get_E_from_U_C_or_E(self,
    #         U=None,
    #         C=None,
    #         E=None):

    #     if (E is not None):
    #         assert (E.ufl_shape[0] == E.ufl_shape[1])
    #     elif (U is not None) or (C is not None):
    #         C = self.get_C_from_U_or_C(U, C)
    #         dim = C.ufl_shape[0]
    #         I = dolfin.Identity(dim)
    #         E = (C - I)/2
    #     else:
    #         assert (0),\
    #             "Must provide U, C or E. Aborting."
    #     return dolfin.variable(E)



    # def get_E_sph_from_U_C_E_or_E_sph(self,
    #         U=None,
    #         C=None,
    #         E=None,
    #         E_sph=None):

    #     if (E_sph is not None):
    #         assert (E_sph.ufl_shape[0] == E_sph.ufl_shape[1])
    #     elif (U is not None) or (C is not None) or (E is not None):
    #         E = self.get_E_from_U_C_or_E(U, C, E)
    #         dim = E.ufl_shape[0]
    #         I = dolfin.Identity(dim)
    #         E_sph = dolfin.tr(E)/dim * I
    #     else:
    #         assert (0),\
    #             "Must provide U, C, E or E_sph. Aborting."
    #     return dolfin.variable(E_sph)



    # def get_E_dev_from_U_C_E_or_E_dev(self,
    #         U=None,
    #         C=None,
    #         E=None,
    #         E_dev=None):

    #     if (E_dev is not None):
    #         assert (E_dev.ufl_shape[0] == E_dev.ufl_shape[1])
    #     elif (U is not None) or (C is not None) or (E is not None):
    #         E = self.get_E_from_U_C_or_E(U, C, E)
    #         E_sph = self.get_E_sph_from_U_C_E_or_E_sph(U, C, E)
    #         E_dev = E - E_sph
    #     else:
    #         assert (0),\
    #             "Must provide U, C, E or E_dev. Aborting."
    #     return dolfin.variable(E_dev)



    # def get_epsilon_from_U_or_epsilon(self,
    #         U=None,
    #         epsilon=None):

    #     if (epsilon is not None):
    #         assert (epsilon.ufl_shape[0] == epsilon.ufl_shape[1])
    #     elif (U is not None):
    #         epsilon = dolfin.sym(dolfin.grad(U))
    #     else:
    #         assert (0),\
    #             "Must provide U or epsilon. Aborting."
    #     return dolfin.variable(epsilon)



    # def get_epsilon_sph_from_U_epsilon_or_epsilon_sph(self,
    #         U=None,
    #         epsilon=None,
    #         epsilon_sph=None):

    #     if (epsilon_sph is not None):
    #         assert (epsilon_sph.ufl_shape[0] == epsilon_sph.ufl_shape[1])
    #     elif (U is not None) or (epsilon is not None):
    #         epsilon = self.get_epsilon_from_U_or_epsilon(U, epsilon)
    #         dim = epsilon.ufl_shape[0]
    #         I = dolfin.Identity(dim)
    #         epsilon_sph = dolfin.tr(epsilon)/dim * I
    #     else:
    #         assert (0),\
    #             "Must provide U, epsilon or epsilon_sph. Aborting."
    #     return dolfin.variable(epsilon_sph)



    # def get_epsilon_dev_from_U_epsilon_or_epsilon_dev(self,
    #         U=None,
    #         epsilon=None,
    #         epsilon_dev=None):

    #     if (epsilon_dev is not None):
    #         assert (epsilon_dev.ufl_shape[0] == epsilon_dev.ufl_shape[1])
    #     elif (U is not None) or (epsilon is not None):
    #         epsilon = self.get_epsilon_from_U_or_epsilon(U, epsilon)
    #         epsilon_sph = self.get_epsilon_sph_from_U_epsilon_or_epsilon_sph(U, epsilon)
    #         epsilon_dev = epsilon - epsilon_sph
    #     else:
    #         assert (0),\
    #             "Must provide U, epsilon or epsilon_dev. Aborting."
    #     return dolfin.variable(epsilon_dev)



################################################################################



def material_factory(
        kinematics,
        model,
        parameters):

    if   (model in ("hooke", "Hooke", "H")):
        material = dmech.HookeElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("hooke_dev", "Hooke_dev", "H_dev")):
        material = dmech.HookeDevElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("hooke_bulk", "Hooke_bulk", "H_bulk")):
        material = dmech.HookeBulkElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("saintvenantkirchhoff", "SaintVenantKirchhoff", "kirchhoff", "Kirchhoff", "SVK")):
        material = dmech.KirchhoffElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("saintvenantkirchhoff_dev", "SaintVenantKirchhoff_dev", "kirchhoff_dev", "Kirchhoff_dev", "SVK_dev")):
        material = dmech.KirchhoffDevElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("saintvenantkirchhoff_bulk", "SaintVenantKirchhoff_bulk", "kirchhoff_bulk", "Kirchhoff_bulk", "SVK_bulk")):
        material = dmech.KirchhoffBulkElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("neohookean", "NeoHookean", "NH")):
        material = dmech.NeoHookeanElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("neohookean_bar", "NeoHookean_bar", "NH_bar")):
        material = dmech.NeoHookeanElasticMaterial(kinematics=kinematics, parameters=parameters, decoup=True)
    elif (model in ("mooneyrivlin", "MooneyRivlin", "MR")):
        material = dmech.MooneyRivlinElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("mooneyrivlin_bar", "MooneyRivlin_bar", "MR_bar")):
        material = dmech.MooneyRivlinElasticMaterial(kinematics=kinematics, parameters=parameters, decoup=True)
    elif (model in ("neohookeanmooneyrivlin", "NeoHookeanMooneyRivlin", "NHMR")):
        material = dmech.NeoHookeanMooneyRivlinElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("neohookeanmooneyrivlin_bar", "NeoHookeanMooneyRivlin_bar", "NHMR_bar")):
        material = dmech.NeoHookeanMooneyRivlinElasticMaterial(kinematics=kinematics, parameters=parameters, decoup=True)
    elif (model in ("ogdenciarletgeymonat", "OgdenCiarletGeymonat", "OCG", "ciarletgeymonat", "CiarletGeymonat", "CG")):
        material = dmech.OgdenCiarletGeymonatElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("ogdenciarletgeymonat_bar", "OgdenCiarletGeymonat_bar", "OCG_bar", "ciarletgeymonat_bar", "CiarletGeymonat_bar", "CG_bar")):
        material = dmech.OgdenCiarletGeymonatElasticMaterial(kinematics=kinematics, parameters=parameters, decoup=True)
    elif (model in ("ogdenciarletgeymonatneohookean", "OgdenCiarletGeymonatNeoHookean", "OCGNH", "ciarletgeymonatneohookean", "CiarletGeymonatNeoHookean", "CGNH")):
        material = dmech.OgdenCiarletGeymonatNeoHookeanElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("ogdenciarletgeymonatneohookean_bar", "OgdenCiarletGeymonatNeoHookean_bar", "OCGNH_bar", "ciarletgeymonatneohookean_bar", "CiarletGeymonatNeoHookean_bar", "CGNH_bar")):
        material = dmech.OgdenCiarletGeymonatNeoHookeanElasticMaterial(kinematics=kinematics, parameters=parameters, decoup=True)
    elif (model in ("ogdenciarletgeymonatneohookeanmooneyrivlin", "OgdenCiarletGeymonatNeoHookeanMooneyRivlin", "OCGNHMR", "ciarletgeymonatneohookeanmooneyrivlin", "CiarletGeymonatNeoHookeanMooneyRivlin", "CGNHMR")):
        material = dmech.OgdenCiarletGeymonatNeoHookeanMooneyRivlinElasticMaterial(kinematics=kinematics, parameters=parameters)
    elif (model in ("ogdenciarletgeymonatneohookeanmooneyrivlin_bar", "OgdenCiarletGeymonatNeoHookeanMooneyRivlin_bar", "OCGNHMR_bar", "ciarletgeymonatneohookeanmooneyrivlin_bar", "CiarletGeymonatNeoHookeanMooneyRivlin_bar", "CGNHMR_bar")):
        material = dmech.OgdenCiarletGeymonatNeoHookeanMooneyRivlinElasticMaterial(kinematics=kinematics, parameters=parameters, decoup=True)
    else:
        assert(0), "Material model (\""+model+"\") not recognized. Aborting."
    return material



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Mahdi Manoochehrtayebi, 2020-2024                                    ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import numpy

import dolfin_mech as dmech
from .Problem_Hyperelasticity import HyperelasticityProblem

################################################################################

class MicroPoroHyperelasticityProblem(HyperelasticityProblem):



    def __init__(self,
            w_solid_incompressibility=False,
            mesh=None,
            mesh_bbox=None,
            vertices=None,
            domains_mf=None,
            boundaries_mf=None,
            points_mf=None,
            displacement_perturbation_degree=None,
            solid_pressure_degree=None,
            quadrature_degree=None,
            foi_degree=0,
            solid_behavior=None,
            bcs="kubc"): # "kubc" or "pbc"

        HyperelasticityProblem.__init__(self)

        self.w_solid_incompressibility = w_solid_incompressibility
        self.vertices = vertices
        if (mesh is not None):
            self.set_mesh(
                mesh=mesh,
                define_spatial_coordinates=1,
                define_facet_normals=1,
                compute_bbox=(mesh_bbox is None))
            self.X_0 = [0.]*self.dim
            # self.X_0 = dolfin.Constant(self.X_0)
            for k_dim in range(self.dim):
                self.X_0[k_dim] = dolfin.assemble(self.X[k_dim] * self.dV)/self.mesh_V0
            self.X_0 = dolfin.Constant(self.X_0)
            if (mesh_bbox is not None):
                self.mesh_bbox = mesh_bbox
            d = [0]*self.dim
            for k_dim in range(self.dim):
                d[k_dim] = self.mesh_bbox[2*k_dim+1] - self.mesh_bbox[2*k_dim+0]

            self.V0 = numpy.prod(d) 
            self.Vs0 = self.mesh_V0
            self.Vf0 = self.V0 - self.Vs0

            self.set_measures(
                domains=domains_mf,
                boundaries=boundaries_mf,
                points=points_mf)

            self.set_subsols(
                displacement_perturbation_degree=displacement_perturbation_degree,
                solid_pressure_degree=solid_pressure_degree)
            self.set_solution_finite_element()
            if (bcs == "pbc"):
                periodic_sd = dmech.PeriodicSubDomain(self.dim, self.mesh_bbox, self.vertices)
                self.set_solution_function_space(constrained_domain=periodic_sd)
            else:
                self.set_solution_function_space()
            self.set_solution_functions()

            self.U_bar      = dolfin.dot(self.get_macroscopic_stretch_subsol().subfunc , self.X-self.X_0)
            self.U_bar_old  = dolfin.dot(self.get_macroscopic_stretch_subsol().func_old, self.X-self.X_0)
            self.U_bar_test = dolfin.dot(self.get_macroscopic_stretch_subsol().dsubtest, self.X-self.X_0)

            self.U_tot      = self.U_bar      + self.get_displacement_perturbation_subsol().subfunc
            self.U_tot_old  = self.U_bar_old  + self.get_displacement_perturbation_subsol().func_old
            self.U_tot_test = self.U_bar_test + self.get_displacement_perturbation_subsol().dsubtest

            self.set_quadrature_degree(
                quadrature_degree=quadrature_degree)

            self.set_foi_finite_elements_DG(
                degree=foi_degree)
            self.set_foi_function_spaces()

            self.add_foi(
                expr=self.U_bar,
                fs=self.get_displacement_perturbation_function_space().collapse(),
                name="U_bar",
                update_type="project")
            self.add_foi(
                expr=self.U_tot,
                fs=self.get_displacement_perturbation_function_space().collapse(),
                name="U_tot",
                update_type="project")

            self.set_kinematics()

            self.add_elasticity_operator(
                solid_behavior_model=solid_behavior["model"],
                solid_behavior_parameters=solid_behavior["parameters"])
            if (self.w_solid_incompressibility):
                self.add_hydrostatic_pressure_operator()
                self.add_incompressibility_operator()

            # self.add_macroscopic_stretch_symmetry_operator()
            self.add_macroscopic_stretch_symmetry_penalty_operator(pen_val=1e6)

            # self.add_deformed_total_volume_operator()
            # self.add_deformed_solid_volume_operator()
            # self.add_deformed_fluid_volume_operator()

            if (bcs == "kubc"):
                self.add_kubc()
            elif (bcs == "pbc"):
                pinpoint_sd = dmech.PinpointSubDomain(coords=mesh.coordinates()[-1], tol=1e-3)
                self.add_constraint(
                    V=self.get_displacement_perturbation_function_space(), 
                    val=[0.]*self.dim,
                    sub_domain=pinpoint_sd,
                    method='pointwise')



    def get_macroscopic_stretch_name(self):

        return "U_bar"

    def add_macroscopic_stretch_subsol(self,
            degree=0,
            symmetry=None,
            init_val=None):

        self.add_tensor_subsol(
            name=self.get_macroscopic_stretch_name(),
            family="R",
            degree=degree,
            symmetry=symmetry,
            init_val=init_val)

    def get_macroscopic_stretch_subsol(self):

        return self.get_subsol(self.get_macroscopic_stretch_name())

    def get_macroscopic_stretch_function_space(self):

        return self.get_subsol_function_space(name=self.get_macroscopic_stretch_name())



    def get_displacement_perturbation_name(self):

        return "U_tilde"

    def add_displacement_perturbation_subsol(self,
            degree):

        self.displacement_perturbation_degree = degree
        self.add_vector_subsol(
            name=self.get_displacement_perturbation_name(),
            family="CG",
            degree=self.displacement_perturbation_degree)

    def get_displacement_perturbation_subsol(self):

        return self.get_subsol(self.get_displacement_perturbation_name())

    def get_displacement_perturbation_function_space(self):

        return self.get_subsol_function_space(name=self.get_displacement_perturbation_name())




    def get_deformed_total_volume_name(self):

        return "v"

    def add_deformed_total_volume_subsol(self):

        self.add_scalar_subsol(
            name=self.get_deformed_total_volume_name(),
            family="R",
            degree=0,
            init_val=self.V0)

    def get_deformed_total_volume_subsol(self):

        return self.get_subsol(self.get_deformed_total_volume_name())



    def get_deformed_solid_volume_name(self):

        return "v_s"

    def add_deformed_solid_volume_subsol(self):

        self.add_scalar_subsol(
            name=self.get_deformed_solid_volume_name(),
            family="R",
            degree=0,
            init_val=self.mesh_V0)

    def get_deformed_solid_volume_subsol(self):

        return self.get_subsol(self.get_deformed_solid_volume_name())



    def get_deformed_fluid_volume_name(self):

        return "v_f"

    def add_deformed_fluid_volume_subsol(self):

        self.add_scalar_subsol(
            name=self.get_deformed_fluid_volume_name(),
            family="R",
            degree=0,
            init_val=self.Vf0)

    def get_deformed_fluid_volume_subsol(self):

        return self.get_subsol(self.get_deformed_fluid_volume_name())



    def set_subsols(self,
            displacement_perturbation_degree=None,
            solid_pressure_degree=None):

        self.add_macroscopic_stretch_subsol(
            symmetry=None) # MG20220425: True does not work, cf. https://fenicsproject.discourse.group/t/writing-symmetric-tensor-function-fails/1136/2 & https://bitbucket.org/fenics-project/dolfin/issues/1065/cannot-store-symmetric-tensor-values

        self.add_displacement_perturbation_subsol(
            degree=displacement_perturbation_degree)
        
        if (self.w_solid_incompressibility):
            if (solid_pressure_degree is None):
                solid_pressure_degree = displacement_perturbation_degree-1
            self.add_pressure_subsol(
                degree=solid_pressure_degree)

        # self.add_macroscopic_stress_lagrange_multiplier_subsol()

        # self.add_deformed_total_volume_subsol()
        # self.add_deformed_solid_volume_subsol()
        # self.add_deformed_fluid_volume_subsol()
        # self.add_surface_area_subsol()



    def set_kinematics(self):

        self.kinematics = dmech.Kinematics(
            U=self.U_tot,
            U_old=self.U_tot_old)

        self.add_foi(expr=self.kinematics.F, fs=self.mfoi_fs, name="F_tot", update_type="project")
        self.add_foi(expr=self.kinematics.J, fs=self.sfoi_fs, name="J_tot", update_type="project")
        self.add_foi(expr=self.kinematics.C, fs=self.mfoi_fs, name="C_tot", update_type="project")
        self.add_foi(expr=self.kinematics.E, fs=self.mfoi_fs, name="E_tot", update_type="project")



    def add_elasticity_operator(self,
            solid_behavior_model,
            solid_behavior_parameters):

        operator = dmech.HyperElasticityOperator(
            U=self.get_displacement_perturbation_subsol().subfunc,
            U_test=self.get_displacement_perturbation_subsol().dsubtest,
            kinematics=self.kinematics,
            material_model=solid_behavior_model,
            material_parameters=solid_behavior_parameters,
            measure=self.dV,
            formulation="ener")
        self.add_foi(expr=operator.material.Sigma, fs=self.mfoi_fs, name="Sigma", update_type="project")
        self.add_foi(expr=operator.material.sigma, fs=self.mfoi_fs, name="sigma", update_type="project")

        return self.add_operator(operator)



    def add_macroscopic_stretch_symmetry_penalty_operator(self,
            **kwargs):

        operator = dmech.MacroscopicStretchSymmetryPenaltyOperator(
            U_bar=self.get_macroscopic_stretch_subsol().subfunc,
            sol=self.sol_func,
            sol_test=self.dsol_test,
            measure=self.dV,
            **kwargs)
        return self.add_operator(operator)



    def add_macroscopic_stretch_component_penalty_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.MacroscopicStretchComponentPenaltyOperator(
            U_bar=self.get_macroscopic_stretch_subsol().subfunc,
            U_bar_test=self.get_macroscopic_stretch_subsol().dsubtest,
            measure=self.dV,
            **kwargs)
        return self.add_operator(operator, k_step=k_step)



    def add_macroscopic_stress_component_constraint_operator(self,
            k_step=None,
            **kwargs):

        for operator in self.operators: # MG20221110: Warning! Only works if there is a single operator with a material law!!
            if hasattr(operator, "material"):
                material = operator.material
                break

        operator = dmech.MacroscopicStressComponentConstraintOperator(
            U_bar=self.get_macroscopic_stretch_subsol().subfunc,
            U_bar_test=self.get_macroscopic_stretch_subsol().dsubtest,
            kinematics=self.kinematics,
            material=material,
            V0=self.V0,
            Vs0=self.Vs0,
            measure=self.dV,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator, k_step=k_step)



    def add_surface_pressure_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.SurfacePressureLoadingOperator(
            U_test=self.get_displacement_perturbation_subsol().dsubtest,
            kinematics=self.kinematics,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_deformed_total_volume_operator(self,
            k_step=None):

        operator = dmech.DeformedTotalVolumeOperator(
            v=self.get_deformed_total_volume_subsol().subfunc,
            v_test=self.get_deformed_total_volume_subsol().dsubtest,
            U_bar=self.get_macroscopic_stretch_subsol().subfunc,
            V0=self.V0,
            measure=self.dV)
        self.add_operator(operator=operator, k_step=k_step)



    def add_deformed_solid_volume_operator(self,
            k_step=None):

        operator = dmech.DeformedSolidVolumeOperator(
            vs=self.get_deformed_solid_volume_subsol().subfunc,
            vs_test=self.get_deformed_solid_volume_subsol().dsubtest,
            J=self.kinematics.J,
            Vs0=self.mesh_V0,
            measure=self.dV)
        self.add_operator(operator=operator, k_step=k_step)



    def add_deformed_fluid_volume_operator(self,
            k_step=None):

        operator = dmech.DeformedFluidVolumeOperator(
            vf=self.get_deformed_fluid_volume_subsol().subfunc,
            vf_test=self.get_deformed_fluid_volume_subsol().dsubtest,
            kinematics=self.kinematics,
            N=self.mesh_normals,
            dS=self.dS,
            U_tot=self.U_tot,
            X=self.X,
            measure=self.dV)
        self.add_operator(operator=operator, k_step=k_step)



    def add_kubc(self,
            xmin_id=1, xmax_id=2,
            ymin_id=3, ymax_id=4,
            zmin_id=5, zmax_id=6):

        self.add_constraint(
            V=self.get_displacement_perturbation_function_space().sub(0),
            sub_domains=self.boundaries,
            sub_domain_id=xmin_id,
            val=0.)
        self.add_constraint(
            V=self.get_displacement_perturbation_function_space().sub(0),
            sub_domains=self.boundaries,
            sub_domain_id=xmax_id,
            val=0.)
        self.add_constraint(
            V=self.get_displacement_perturbation_function_space().sub(1),
            sub_domains=self.boundaries,
            sub_domain_id=ymin_id,
            val=0.)
        self.add_constraint(
            V=self.get_displacement_perturbation_function_space().sub(1),
            sub_domains=self.boundaries,
            sub_domain_id=ymax_id,
            val=0.)
        if (self.dim==3):
            self.add_constraint(
                V=self.get_displacement_perturbation_function_space().sub(2),
                sub_domains=self.boundaries,
                sub_domain_id=zmin_id,
                val=0.)
            self.add_constraint(
                V=self.get_displacement_perturbation_function_space().sub(2),
                sub_domains=self.boundaries,
                sub_domain_id=zmax_id,
                val=0.)



    def add_deformed_solid_volume_qoi(self):

        self.add_qoi(
            name="vs",
            expr=self.kinematics.J * self.dV)



    def add_deformed_fluid_volume_qoi(self):

        U_bar = self.get_macroscopic_stretch_subsol().subfunc
        I_bar = dolfin.Identity(self.dim)
        F_bar = I_bar + U_bar
        J_bar = dolfin.det(F_bar)
        v = J_bar * self.V0

        self.add_qoi(
            name="vf",
            expr=(v/self.Vs0 - self.kinematics.J) * self.dV)



    def add_deformed_volume_qoi(self):

        U_bar = self.get_macroscopic_stretch_subsol().subfunc
        I_bar = dolfin.Identity(self.dim)
        F_bar = I_bar + U_bar
        J_bar = dolfin.det(F_bar)
        v = J_bar * self.V0

        self.add_qoi(
            name="v",
            expr=(v/self.Vs0) * self.dV)



    def add_macroscopic_tensor_qois(self,
            basename,
            get_subsol,
            symmetric=False):

        self.add_qoi(
            name=basename+"_XX",
            expr=get_subsol().subfunc[0,0],
            point=self.mesh.coordinates()[0],
            update_type="direct")
        if (self.dim >= 2):
            self.add_qoi(
                name=basename+"_YY",
                expr=get_subsol().subfunc[1,1],
                point=self.mesh.coordinates()[0],
                update_type="direct")
            if (self.dim >= 3):
                self.add_qoi(
                    name=basename+"_ZZ",
                    expr=get_subsol().subfunc[2,2],
                    point=self.mesh.coordinates()[0],
                    update_type="direct")
        if (self.dim >= 2):
            self.add_qoi(
                name=basename+"_XY",
                expr=get_subsol().subfunc[0,1],
                point=self.mesh.coordinates()[0],
                update_type="direct")
            if not (symmetric): self.add_qoi(
                name=basename+"_YX",
                expr=get_subsol().subfunc[1,0],
                point=self.mesh.coordinates()[0],
                update_type="direct")
            if (self.dim >= 3):
                self.add_qoi(
                    name=basename+"_YZ",
                    expr=get_subsol().subfunc[1,2],
                    point=self.mesh.coordinates()[0],
                    update_type="direct")
                if not (symmetric): self.add_qoi(
                    name=basename+"_ZY",
                    expr=get_subsol().subfunc[2,1],
                    point=self.mesh.coordinates()[0],
                    update_type="direct")
                self.add_qoi(
                    name=basename+"_ZX",
                    expr=get_subsol().subfunc[2,0],
                    point=self.mesh.coordinates()[0],
                    update_type="direct")
                if not (symmetric): self.add_qoi(
                    name=basename+"_XZ",
                    expr=get_subsol().subfunc[0,2],
                    point=self.mesh.coordinates()[0],
                    update_type="direct")



    def add_macroscopic_stretch_qois(self):

        self.add_macroscopic_tensor_qois(
            basename="U_bar",
            get_subsol=self.get_macroscopic_stretch_subsol)



    def add_macroscopic_solid_stress_qois(self,
            symmetric=False):

        for operator in self.operators: # MG20221110: Warning! Only works if there is a single operator with a material law!!
            if hasattr(operator, "material"):
                material = operator.material
                break

        U_bar = self.get_macroscopic_stretch_subsol().subfunc
        I_bar = dolfin.Identity(self.dim)
        F_bar = I_bar + U_bar
        J_bar = dolfin.det(F_bar)
        v = J_bar * self.V0

        self.add_qoi(
            name="sigma_s_bar_XX",
            expr=(material.sigma[0,0] * self.kinematics.J)/v * self.dV)
        if (self.dim >= 2):
            self.add_qoi(
                name="sigma_s_bar_YY",
                expr=(material.sigma[1,1] * self.kinematics.J)/v * self.dV)
            if (self.dim >= 3):
                self.add_qoi(
                    name="sigma_s_bar_ZZ",
                    expr=(material.sigma[2,2] * self.kinematics.J )/v * self.dV)
        if (self.dim >= 2):
            self.add_qoi(
                name="sigma_s_bar_XY",
                expr=(material.sigma[0,1] * self.kinematics.J)/v * self.dV)
            if not (symmetric): self.add_qoi(
                name="sigma_s_bar_YX",
                expr=(material.sigma[1,0] * self.kinematics.J)/v * self.dV)
            if (self.dim >= 3):
                self.add_qoi(
                    name="sigma_s_bar_YZ",
                    expr=(material.sigma[1,2] * self.kinematics.J)/v * self.dV)
                if not (symmetric): self.add_qoi(
                    name="sigma_s_bar_ZY",
                    expr=(material.sigma[2,1] * self.kinematics.J)/v * self.dV)
                self.add_qoi(
                    name="sigma_s_bar_ZX",
                    expr=(material.sigma[2,0] * self.kinematics.J)/v * self.dV)
                if not (symmetric): self.add_qoi(
                    name="sigma_s_bar_XZ",
                    expr=(material.sigma[0,2] * self.kinematics.J)/v * self.dV)



    def add_macroscopic_solid_hydrostatic_pressure_qoi(self):

        for operator in self.operators: # MG20221110: Warning! Only works if there is a single operator with a material law!!
            if hasattr(operator, "material"):
                material = operator.material
                break

        U_bar = self.get_macroscopic_stretch_subsol().subfunc
        I_bar = dolfin.Identity(self.dim)
        F_bar = I_bar + U_bar
        J_bar = dolfin.det(F_bar)
        v = J_bar * self.V0

        self.add_qoi(
            name="p_hydro",
            expr=(material.p_hydro * self.kinematics.J)/v * self.dV)



    def add_fluid_pressure_qoi(self,
            symmetric=False):

        for operator in self.steps[0].operators: # MG20231124: Warning! Only works if there is a single step!!
            if hasattr(operator, "tv_pf"):
                tv_pf = operator.tv_pf
                break

        self.add_qoi(
            name="p_f",
            expr=(tv_pf.val)/self.Vs0 * self.dV)



    def add_macroscopic_stress_qois(self,
            symmetric=False):

        for operator in self.operators: # MG20221110: Warning! Only works if there is a single operator with a material law!!
            if hasattr(operator, "material"):
                material = operator.material
                break

        for operator in self.steps[0].operators: # MG20231124: Warning! Only works if there is a single step!!
            if hasattr(operator, "tv_pf"):
                tv_pf = operator.tv_pf
                break

        U_bar = self.get_macroscopic_stretch_subsol().subfunc
        I_bar = dolfin.Identity(self.dim)
        F_bar = I_bar + U_bar
        J_bar = dolfin.det(F_bar)
        v = J_bar * self.V0

        self.add_qoi(
            name="sigma_bar_XX",
            expr=(material.sigma[0,0] * self.kinematics.J - (v/self.Vs0 - self.kinematics.J) * tv_pf.val)/v * self.dV)
        if (self.dim >= 2):
            self.add_qoi(
                name="sigma_bar_YY",
                expr=(material.sigma[1,1] * self.kinematics.J - (v/self.Vs0 - self.kinematics.J) * tv_pf.val)/v * self.dV)
            if (self.dim >= 3):
                self.add_qoi(
                    name="sigma_bar_ZZ",
                    expr=(material.sigma[2,2] * self.kinematics.J - (v/self.Vs0 - self.kinematics.J) * tv_pf.val)/v * self.dV)
        if (self.dim >= 2):
            self.add_qoi(
                name="sigma_bar_XY",
                expr=(material.sigma[0,1] * self.kinematics.J)/v * self.dV)
            if not (symmetric): self.add_qoi(
                name="sigma_bar_YX",
                expr=(material.sigma[1,0] * self.kinematics.J)/v * self.dV)
            if (self.dim >= 3):
                self.add_qoi(
                    name="sigma_bar_YZ",
                    expr=(material.sigma[1,2] * self.kinematics.J)/v * self.dV)
                if not (symmetric): self.add_qoi(
                    name="sigma_bar_ZY",
                    expr=(material.sigma[2,1] * self.kinematics.J)/v * self.dV)
                self.add_qoi(
                    name="sigma_bar_ZX",
                    expr=(material.sigma[2,0] * self.kinematics.J)/v * self.dV)
                if not (symmetric): self.add_qoi(
                    name="sigma_bar_XZ",
                    expr=(material.sigma[0,2] * self.kinematics.J)/v * self.dV)






#coding=utf8

################################################################################
###                                                                          ###
### Created by Mahdi Manoochehrtayebi, 2020-2024                             ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Martin Genet, 2018-2024                                              ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class MacroscopicStressComponentConstraintOperator(Operator):

    def __init__(self,
            U_bar, U_bar_test,
            kinematics,
            material,
            V0, Vs0,
            i, j,
            measure,
            N,
            sigma_bar_ij_val=None, sigma_bar_ij_ini=None, sigma_bar_ij_fin=None,
            pf_val=None, pf_ini=None, pf_fin=None):

        self.kinematics = kinematics
        self.material = material
        self.measure  = measure
        self.N = N

        self.tv_pf = dmech.TimeVaryingConstant(
            val=pf_val, val_ini=pf_ini, val_fin=pf_fin)
        pf = self.tv_pf.val

        self.tv_sigma_bar_ij = dmech.TimeVaryingConstant(
            val=sigma_bar_ij_val, val_ini=sigma_bar_ij_ini, val_fin=sigma_bar_ij_fin)
        sigma_bar_ij = self.tv_sigma_bar_ij.val

        dim = U_bar.ufl_shape[0]
        I_bar = dolfin.Identity(dim)
        F_bar = I_bar + U_bar
        J_bar = dolfin.det(F_bar)
        v = J_bar * V0

        sigma_tilde = self.material.sigma * self.kinematics.J - (v/Vs0 - self.kinematics.J) * pf * I_bar
        self.res_form = U_bar_test[i,j] * (sigma_tilde[i,j] - (v/Vs0) * sigma_bar_ij) * self.measure



    def set_value_at_t_step(self,
            t_step):

        self.tv_pf.set_value_at_t_step(t_step)
        self.tv_sigma_bar_ij.set_value_at_t_step(t_step)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class HookeElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters):

        self.kinematics = kinematics

        self.lmbda = self.get_lambda_from_parameters(parameters)
        self.mu    = self.get_mu_from_parameters(parameters)

        self.psi   = (self.lmbda/2) * dolfin.tr(self.kinematics.epsilon)**2 + self.mu * dolfin.inner(self.kinematics.epsilon, self.kinematics.epsilon)

        self.sigma = dolfin.diff(self.psi, self.kinematics.epsilon)
        # self.sigma = self.lmbda * dolfin.tr(self.kinematics.epsilon) * self.kinematics.I + 2 * self.mu * self.kinematics.epsilon

        self.Sigma = self.sigma
        self.P     = self.sigma

        if (self.kinematics.dim == 2):
            self.sigma_ZZ = self.lmbda * dolfin.tr(self.kinematics.epsilon)



    # def get_free_energy(self,
    #         U=None,
    #         epsilon=None):

    #     epsilon = self.get_epsilon_from_U_or_epsilon(
    #         U, epsilon)

    #     psi = (self.lmbda/2) * dolfin.tr(epsilon)**2 + self.mu * dolfin.inner(epsilon, epsilon)
    #     sigma = dolfin.diff(psi, epsilon)

    #     # assert (epsilon.ufl_shape[0] == epsilon.ufl_shape[1])
    #     # dim = epsilon.ufl_shape[0]
    #     # I = dolfin.Identity(dim)
    #     # sigma = self.lmbda * dolfin.tr(epsilon) * I + 2 * self.mu * epsilon

    #     return psi, sigma

################################################################################

class HookeBulkElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters):

        self.kinematics = kinematics

        # self.K = self.get_K_from_parameters(parameters)
        self.lmbda, self.mu = self.get_lambda_and_mu_from_parameters(parameters)
        self.K = (self.kinematics.dim*self.lmbda + 2*self.mu)/self.kinematics.dim

        self.psi   = (self.kinematics.dim*self.K/2) * dolfin.tr(self.kinematics.epsilon_sph)**2
        self.sigma =  self.kinematics.dim*self.K    *           self.kinematics.epsilon_sph

        self.Sigma = self.sigma
        self.P     = self.sigma

        if (self.kinematics.dim == 2):
            self.sigma_ZZ = self.K * dolfin.tr(self.kinematics.epsilon)



    # def get_free_energy(self,
    #         U=None,
    #         epsilon=None,
    #         epsilon_sph=None):

    #     epsilon_sph = self.get_epsilon_sph_from_U_epsilon_or_epsilon_sph(
    #         U, epsilon, epsilon_sph)
    #     assert (epsilon_sph.ufl_shape[0] == epsilon_sph.ufl_shape[1])
    #     dim = epsilon_sph.ufl_shape[0]

    #     psi   = (dim*self.K/2) * dolfin.tr(epsilon_sph)**2
    #     sigma =  dim*self.K    *           epsilon_sph

    #     return psi, sigma



    # def get_Cauchy_stress(self,
    #         U=None,
    #         epsilon=None,
    #         epsilon_sph=None):

    #     epsilon_sph = self.get_epsilon_sph_from_U_epsilon_or_epsilon_sph(
    #         U, epsilon, epsilon_sph)
    #     assert (epsilon_sph.ufl_shape[0] == epsilon_sph.ufl_shape[1])
    #     dim = epsilon_sph.ufl_shape[0]

    #     sigma = dim * self.K * epsilon_sph

    #     return sigma

################################################################################

class HookeDevElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters):

        self.kinematics = kinematics

        self.G = self.get_G_from_parameters(parameters)

        self.psi   =   self.G * dolfin.inner(self.kinematics.epsilon_dev, self.kinematics.epsilon_dev)
        self.sigma = 2*self.G *              self.kinematics.epsilon_dev

        self.Sigma = self.sigma
        self.P     = self.sigma

        if (self.kinematics.dim == 2):
            self.sigma_ZZ = -2*self.G/3 * dolfin.tr(self.kinematics.epsilon)



    # def get_free_energy(self,
    #         U=None,
    #         epsilon=None,
    #         epsilon_dev=None):

    #     epsilon_dev = self.get_epsilon_dev_from_U_epsilon_or_epsilon_dev(
    #         U, epsilon, epsilon_dev)
        
    #     psi   =   self.G * dolfin.inner(epsilon_dev, epsilon_dev)
    #     sigma = 2*self.G *              epsilon_dev

    #     return psi, sigma



    # def get_Cauchy_stress(self,
    #         U=None,
    #         epsilon=None,
    #         epsilon_dev=None):

    #     epsilon_dev = self.get_epsilon_dev_from_U_epsilon_or_epsilon_dev(
    #         U, epsilon, epsilon_dev)
        
    #     sigma = 2*self.G * epsilon_dev

    #     return sigma



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class LinearizedHydrostaticPressureOperator(Operator):

    def __init__(self,
            kinematics,
            u_test,
            p,
            measure):

        self.kinematics = kinematics
        self.p          = p
        self.measure    = measure

        epsilon_test = dolfin.sym(dolfin.grad(u_test))
        self.res_form = -self.p * dolfin.tr(epsilon_test) * self.measure



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class SurfacePressureLoadingOperator(Operator):

    def __init__(self,
            U_test,
            kinematics,
            N,
            measure,
            P_val=None, P_ini=None, P_fin=None):

        self.measure = measure

        self.tv_P = dmech.TimeVaryingConstant(
            val=P_val, val_ini=P_ini, val_fin=P_fin)
        P = self.tv_P.val

        T = dolfin.dot(-P * N, dolfin.inv(kinematics.F))
        self.res_form = - dolfin.inner(T, U_test) * kinematics.J * self.measure



    def set_value_at_t_step(self,
            t_step):

        self.tv_P.set_value_at_t_step(t_step)

################################################################################

class SurfacePressure0LoadingOperator(Operator):

    def __init__(self,
            U_test,
            N,
            measure,
            P_val=None, P_ini=None, P_fin=None):

        self.measure = measure

        self.tv_P = dmech.TimeVaryingConstant(
            val=P_val, val_ini=P_ini, val_fin=P_fin)
        P = self.tv_P.val

        self.res_form = - dolfin.inner(-P * N, U_test) * self.measure



    def set_value_at_t_step(self,
            t_step):

        self.tv_P.set_value_at_t_step(t_step)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

################################################################################

class XDMFFile():

    def __init__(self,
            filename,
            functions):

        self.xdmf_file = dolfin.XDMFFile(filename)
        self.xdmf_file.parameters["flush_output"] = True
        self.xdmf_file.parameters["functions_share_mesh"] = True
        #self.xdmf_file.parameters["rewrite_function_mesh"] = False

        self.functions = functions



    def close(self):

        self.xdmf_file.close()



    def write(self,
            time=0):

        for function in self.functions:
            self.xdmf_file.write(function, float(time))



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin_mech as dmech
from .Problem import Problem

################################################################################

class ElasticityProblem(Problem):



    def __init__(self,
            w_incompressibility=False,
            mesh=None,
            define_facet_normals=False,
            domains_mf=None,
            boundaries_mf=None,
            points_mf=None,
            displacement_degree=None,
            pressure_degree=None,
            quadrature_degree=None,
            foi_degree=0,
            elastic_behavior=None,
            elastic_behaviors=None):

        Problem.__init__(self)

        self.w_incompressibility = w_incompressibility

        if (mesh is not None):
            self.set_mesh(
                mesh=mesh,
                define_facet_normals=define_facet_normals)

            self.set_measures(
                domains=domains_mf,
                boundaries=boundaries_mf,
                points=points_mf)

            self.set_subsols(
                displacement_degree=displacement_degree,
                pressure_degree=pressure_degree)
            self.set_solution_finite_element()
            self.set_solution_function_space()
            self.set_solution_functions()

            self.set_quadrature_degree(
                quadrature_degree=quadrature_degree)

            self.set_foi_finite_elements_DG(
                degree=foi_degree)
            self.set_foi_function_spaces()

            self.set_kinematics()

            if (elastic_behavior is not None):
                elastic_behaviors = [elastic_behavior]

            self.add_elasticity_operators(
                elastic_behaviors=elastic_behaviors)



    def get_displacement_name(self):
        return "u"



    def add_displacement_subsol(self,
            degree):

        self.displacement_degree = degree
        self.add_vector_subsol(
            name=self.get_displacement_name(),
            family="CG",
            degree=self.displacement_degree)



    def get_displacement_subsol(self):

        return self.get_subsol(self.get_displacement_name())



    def get_displacement_function_space(self):

        if (len(self.subsols) == 1):
            return self.sol_fs
        else:
            return self.get_subsol_function_space(name=self.get_displacement_name())



    def get_pressure_name(self):
        return "p"



    def add_pressure_subsol(self,
            degree):

        self.pressure_degree = degree
        if (self.pressure_degree == 0):
            self.add_scalar_subsol(
                name=self.get_pressure_name(),
                family="DG",
                degree=self.pressure_degree)
        else:
            self.add_scalar_subsol(
                name=self.get_pressure_name(),
                family="CG",
                degree=self.pressure_degree)



    def get_pressure_subsol(self):

        assert (self.w_incompressibility),\
            "There is no pressure subsol. Aborting."
        return self.get_subsol(self.get_pressure_name())



    def get_pressure_function_space(self):

        assert (self.w_incompressibility),\
            "There is no pressure function space. Aborting."
        return self.get_subsol_function_space(name=self.get_pressure_name())



    def set_subsols(self,
            displacement_degree=1,
            pressure_degree=None):

        self.add_displacement_subsol(
            degree=displacement_degree)

        if (self.w_incompressibility):
            if (pressure_degree is None):
                pressure_degree = displacement_degree-1
            self.add_pressure_subsol(
                degree=pressure_degree)



    def set_quadrature_degree(self,
            quadrature_degree=None):

        if (quadrature_degree is None) or (type(quadrature_degree) == int):
            pass
        elif (quadrature_degree == "full"):
            quadrature_degree = None
        elif (quadrature_degree == "default"):
            if   (self.mesh.ufl_cell().cellname() in ("triangle", "tetrahedron")):
                quadrature_degree = max(2, 2*(self.displacement_degree-1)) # MG20211221: This does not allow to reproduce full integration results exactly, but it is quite close…
            elif (self.mesh.ufl_cell().cellname() in ("quadrilateral", "hexahedron")):
                quadrature_degree = max(2, 2*(self.dim*self.displacement_degree-1))
        else:
            assert (0),\
                "Must provide an int, \"full\", \"default\" or None. Aborting."

        Problem.set_quadrature_degree(self,
            quadrature_degree=quadrature_degree)



    def set_kinematics(self):

        self.kinematics = dmech.LinearizedKinematics(
            u=self.get_displacement_subsol().subfunc,
            u_old=self.get_displacement_subsol().func_old)

        self.add_foi(expr=self.kinematics.epsilon, fs=self.mfoi_fs, name="epsilon")



    def get_subdomain_measure(self,
            subdomain_id=None):

        if (subdomain_id is None):
            return self.dV
        else:
            return self.dV(subdomain_id)



    def add_elasticity_operator(self,
            material_model,
            material_parameters,
            subdomain_id=None):

        operator = dmech.LinearizedElasticityOperator(
            kinematics=self.kinematics,
            u_test=self.get_displacement_subsol().dsubtest,
            material_model=material_model,
            material_parameters=material_parameters,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



    def add_hydrostatic_pressure_operator(self,
            subdomain_id=None):

        operator = dmech.LinearizedHydrostaticPressureOperator(
            kinematics=self.kinematics,
            u_test=self.get_displacement_subsol().dsubtest,
            p=self.get_pressure_subsol().subfunc,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



    def add_incompressibility_operator(self,
            subdomain_id=None):

        operator = dmech.LinearizedIncompressibilityOperator(
            kinematics=self.kinematics,
            p_test=self.get_pressure_subsol().dsubtest,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



    def add_elasticity_operators(self,
            elastic_behaviors):

        for elastic_behavior in elastic_behaviors:
            operator = self.add_elasticity_operator(
                material_model=elastic_behavior["model"],
                material_parameters=elastic_behavior["parameters"],
                subdomain_id=elastic_behavior.get("subdomain_id", None))
            suffix = "_"+elastic_behavior["suffix"] if "suffix" in elastic_behavior else ""
            self.add_foi(expr=operator.material.sigma, fs=self.mfoi_fs, name="sigma"+suffix)
        if (self.w_incompressibility):
            self.add_hydrostatic_pressure_operator()
            self.add_incompressibility_operator()



    def add_global_strain_qois(self):

        basename = "e_"
        strain = self.kinematics.epsilon

        self.add_qoi(
            name=basename+"XX",
            expr=strain[0,0] * self.dV)
        if (self.dim >= 2):
            self.add_qoi(
                name=basename+"YY",
                expr=strain[1,1] * self.dV)
            if (self.dim >= 3):
                self.add_qoi(
                    name=basename+"ZZ",
                    expr=strain[2,2] * self.dV)
        if (self.dim >= 2):
            self.add_qoi(
                name=basename+"XY",
                expr=strain[0,1] * self.dV)
            if (self.dim >= 3):
                self.add_qoi(
                    name=basename+"YZ",
                    expr=strain[1,2] * self.dV)
                self.add_qoi(
                    name=basename+"ZX",
                    expr=strain[2,0] * self.dV)



    def add_global_stress_qois(self):

        basename = "s_"

        self.add_qoi(
            name=basename+"XX",
            expr=sum([getattr(operator.material, "sigma")[0,0]*operator.measure
                      for operator in self.operators
                      if (hasattr(operator, "material")
                      and hasattr(operator.material, "sigma"))]))
        if (self.dim >= 2):
            self.add_qoi(
                name=basename+"YY",
                expr=sum([getattr(operator.material, "sigma")[1,1]*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "sigma"))]))
            if (self.dim >= 3):
                self.add_qoi(
                    name=basename+"ZZ",
                    expr=sum([getattr(operator.material, "sigma")[2,2]*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "sigma"))]))
        if (self.dim >= 2):
            self.add_qoi(
                name=basename+"XY",
                expr=sum([getattr(operator.material, "sigma")[0,1]*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "sigma"))]))
            if (self.dim >= 3):
                self.add_qoi(
                    name=basename+"YZ",
                    expr=sum([getattr(operator.material, "sigma")[1,2]*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "sigma"))]))
                self.add_qoi(
                    name=basename+"ZX",
                    expr=sum([getattr(operator.material, "sigma")[2,0]*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "sigma"))]))



    def add_global_pressure_qoi(self):

        self.add_qoi(
            name="p",
            expr=sum([operator.p*operator.measure for operator in self.operators if hasattr(operator, "p")]))
            # expr=sum([-dolfin.tr(operator.material.sigma)/3*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "sigma"))])+sum([operator.p*operator.measure for operator in self.operators if hasattr(operator, "p")]))



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class WporePoroOperator(Operator):

    def __init__(self,
            kinematics,
            Phis0,
            Phis,
            Phis_test,
            material_parameters,
            material_scaling,
            measure):

        self.kinematics = kinematics
        self.solid_material = dmech.WporeLungElasticMaterial(
            Phif=self.kinematics.J - Phis,
            Phif0=1-Phis0,
            parameters=material_parameters)
        self.material = dmech.PorousElasticMaterial(
            solid_material=self.solid_material,
            scaling=material_scaling,
            Phis0=Phis0)
        self.measure = measure

        self.res_form = - self.material.dWporedPhif * Phis_test * self.measure

################################################################################

class InverseWporePoroOperator(Operator):

    def __init__(self,
            kinematics,
            phis,
            phis0,
            phis0_test,
            material_parameters,
            material_scaling,
            measure):

        self.kinematics = kinematics
        self.solid_material = dmech.WporeLungElasticMaterial(
            Phif=self.kinematics.J * (1 - phis),
            Phif0=1 - self.kinematics.J * phis0,
            parameters=material_parameters)
        self.material = dmech.PorousElasticMaterial(
            solid_material=self.solid_material,
            scaling=material_scaling,
            Phis0=self.kinematics.J * phis0)
        self.measure = measure

        self.res_form = - self.material.dWporedPhif * phis0_test * self.measure



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin_mech as dmech
from .Problem_Hyperelasticity import HyperelasticityProblem

################################################################################

class InverseHyperelasticityProblem(HyperelasticityProblem):



    def __init__(self, *args, **kwargs):

        if ("w_incompressibility" in kwargs):
            assert (bool(kwargs["w_incompressibility"]) == 0),\
                "Incompressibility not implemented for inverse problem. Aborting."

        HyperelasticityProblem.__init__(self, *args, **kwargs)



    def get_displacement_name(self):
        return "u"



    def get_pressure_name(self):
        return "p"



    def set_kinematics(self):

        self.kinematics = dmech.InverseKinematics(
            u=self.get_displacement_subsol().subfunc,
            u_old=self.get_displacement_subsol().func_old)

        self.add_foi(expr=self.kinematics.F, fs=self.mfoi_fs, name="F")
        self.add_foi(expr=self.kinematics.J, fs=self.sfoi_fs, name="J")
        self.add_foi(expr=self.kinematics.C, fs=self.mfoi_fs, name="C")
        self.add_foi(expr=self.kinematics.E, fs=self.mfoi_fs, name="E")



    def add_elasticity_operator(self,
            material_model,
            material_parameters,
            subdomain_id=None):

        operator = dmech.LinearizedElasticityOperator(
            kinematics=self.kinematics,
            u_test=self.get_displacement_subsol().dsubtest,
            material_model=material_model,
            material_parameters=material_parameters,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import numpy

import dolfin_mech as dmech

################################################################################

class InverseKinematics():



    def __init__(self,
            u,
            u_old=None):

        self.u = u

        self.dim = self.u.ufl_shape[0]
        self.I = dolfin.Identity(self.dim)

        self.f     = self.I + dolfin.grad(self.u)
        self.F     = dolfin.inv(self.f)
        self.F     = dolfin.variable(self.F)
        self.J     = dolfin.det(self.F)
        self.C     = self.F.T * self.F
        self.C     = dolfin.variable(self.C)
        self.C_inv = dolfin.inv(self.C)
        self.IC    = dolfin.tr(self.C)
        self.IIC   = (dolfin.tr(self.C)*dolfin.tr(self.C) - dolfin.tr(self.C*self.C))/2
        self.E     = (self.C - self.I)/2
        self.E     = dolfin.variable(self.E)

        self.F_bar   = self.J**(-1/self.dim) * self.F
        self.C_bar   = self.F_bar.T * self.F_bar
        self.IC_bar  = dolfin.tr(self.C_bar)
        self.IIC_bar = (dolfin.tr(self.C_bar)*dolfin.tr(self.C_bar) - dolfin.tr(self.C_bar*self.C_bar))/2
        self.E_bar   = (self.C_bar - self.I)/2

        if (u_old is not None):
            self.u_old = u_old

            self.f_old = self.I + dolfin.grad(self.u_old)
            self.F_old = dolfin.inv(self.f_old)
            self.J_old = dolfin.det(self.F_old)
            self.C_old = self.F_old.T * self.F_old
            self.E_old = (self.C_old - self.I)/2

            self.F_bar_old = self.J_old**(-1/self.dim) * self.F_old
            self.C_bar_old = self.F_bar_old.T * self.F_bar_old
            self.E_bar_old = (self.C_bar_old - self.I)/2



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import time

import dolfin_mech as dmech

################################################################################

class FOI():



    def __init__(self,
            expr=None,
            fs=None,
            func=None,
            name=None,
            form_compiler_parameters={},
            update_type="local_solver"): # local_solver or project or interpolate

        if (expr is not None) and (fs is not None):

            self.expr = expr
            self.fs = fs
            self.func = func if func is not None else dolfin.Function(fs)
            if (name is not None):
                self.name = name
                self.func.rename(self.name, self.name)

            if (update_type == "local_solver"):

                self.form_compiler_parameters = form_compiler_parameters

                self.func_test = dolfin.TestFunction(self.fs)
                self.func_tria = dolfin.TrialFunction(self.fs)

                self.a_expr = dolfin.inner(
                    self.func_tria,
                    self.func_test) * dolfin.dx(metadata=self.form_compiler_parameters)
                self.b_expr = dolfin.inner(
                    self.expr,
                    self.func_test) * dolfin.dx(metadata=self.form_compiler_parameters)
                self.local_solver = dolfin.LocalSolver(
                    self.a_expr,
                    self.b_expr)
                # t = time.time()
                self.local_solver.factorize()
                # t = time.time() - t
                # print("LocalSolver factorization = "+str(t)+" s")

                self.update = self.update_local_solver

            elif (update_type == "project"):

                self.form_compiler_parameters = form_compiler_parameters

                self.update = self.update_project

            elif (update_type == "interpolate"):

                self.update = self.update_interpolate

        elif (expr is None) and (fs is None) and (func is not None):

            self.func = func

            self.update = self.update_none



    def update_local_solver(self):

        # print(self.name)
        # print(self.form_compiler_parameters)

        # t = time.time()
        self.local_solver.solve_local_rhs(self.func)
        # t = time.time() - t
        # print("LocalSolver solve = "+str(t)+" s")



    def update_project(self):

        # print(self.name)
        # print(self.form_compiler_parameters)

        # t = time.time()
        dolfin.project(
            v=self.expr,
            V=self.fs,
            function=self.func,
            form_compiler_parameters=self.form_compiler_parameters)
        # t = time.time() - t
        # print("Project = "+str(t)+" s")



    def update_interpolate(self):

        # print(self.name)

        # t = time.time()
        self.func.interpolate(self.expr)
        # t = time.time() - t
        # print("Projec = "+str(t)+" s")



    def update_none(self):

        pass



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class HyperHydrostaticPressureOperator(Operator):

    def __init__(self,
            kinematics,
            U_test,
            P,
            measure):

        self.kinematics = kinematics
        self.P          = P
        self.measure    = measure

        dJ_test = dolfin.derivative(
            self.kinematics.J,
            self.kinematics.U,
            U_test)

        self.res_form = - self.P * dJ_test * self.measure



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

from curses import use_default_colors
import dolfin

import dolfin_mech as dmech

################################################################################

def run_RivlinCube_Hyperelasticity(
        dim                                    : int  = 3                               ,
        inverse                                : bool = 0                               ,
        incomp                                 : bool = 0                               ,
        multimaterial                          : bool = 0                               ,
        cube_params                            : dict = {}                              ,
        mat_params                             : dict = {}                              ,
        step_params                            : dict = {}                              ,
        const_params                           : dict = {}                              ,
        load_params                            : dict = {}                              ,
        move_params                            : dict = {}                              ,
        get_results                            : bool = 0                               ,
        res_basename                           : str  = "run_RivlinCube_Hyperelasticity",
        write_vtus_with_preserved_connectivity : bool = False                           ,
        verbose                                : bool = 0                               ):

    ################################################################### Mesh ###

    if   (dim==2):
        mesh, boundaries_mf, xmin_id, xmax_id, ymin_id, ymax_id = dmech.run_RivlinCube_Mesh(dim=dim, params=cube_params)
    elif (dim==3):
        mesh, boundaries_mf, xmin_id, xmax_id, ymin_id, ymax_id, zmin_id, zmax_id = dmech.run_RivlinCube_Mesh(dim=dim, params=cube_params)

    if move_params.get("move", False) == True :
        Umove = move_params.get("U")
        dolfin.ALE.move(mesh, Umove)

    if (multimaterial):
        mat1_sd = dolfin.CompiledSubDomain("x[0] <= x0", x0=0.5)
        mat2_sd = dolfin.CompiledSubDomain("x[0] >= x0", x0=0.5)

        mat1_id = 1
        mat2_id = 2

        domains_mf = dolfin.MeshFunction("size_t", mesh, mesh.topology().dim()) # MG20180418: size_t looks like unisgned int, but more robust wrt architecture and os
        domains_mf.set_all(0)
        mat1_sd.mark(domains_mf, mat1_id)
        mat2_sd.mark(domains_mf, mat2_id)
    else:
        domains_mf = None

    ################################################################ Problem ###

    if (inverse):
        problem_type = dmech.InverseHyperelasticityProblem
    else:
        problem_type = dmech.HyperelasticityProblem

    if (incomp):
        displacement_degree = 2 # MG20211219: Incompressibility requires displacement_degree >= 2 ?!
        w_incompressibility = 1
    else:
        displacement_degree = 1
        w_incompressibility = 0

    quadrature_degree = "default"
    # quadrature_degree = "full"

    if (multimaterial):
        elastic_behavior = None
        if (incomp):
            mat1_mod = "NHMR"
            mat2_mod = "NHMR"
        else:
            mat1_mod = "CGNHMR"
            mat2_mod = "CGNHMR"
        mat1_params = {
            "E":1.,
            "nu":0.5*(incomp)+0.3*(1-incomp)}

        mat2_params = {
            "E":10.,
            "nu":0.5*(incomp)+0.3*(1-incomp)}
        elastic_behaviors=[
                {"subdomain_id":mat1_id, "model":mat1_mod, "parameters":mat1_params, "suffix":"1"},
                {"subdomain_id":mat2_id, "model":mat2_mod, "parameters":mat2_params, "suffix":"2"}]
    else:
        elastic_behavior = mat_params
        elastic_behaviors = None

    problem = problem_type(
        mesh=mesh,
        domains_mf=domains_mf,
        define_facet_normals=1,
        boundaries_mf=boundaries_mf,
        displacement_degree=displacement_degree,
        quadrature_degree=quadrature_degree,
        w_incompressibility=w_incompressibility,
        elastic_behavior=elastic_behavior,
        elastic_behaviors=elastic_behaviors)

    ########################################## Boundary conditions & Loading ###

    const_type = const_params.get("type", "sym")

    if (const_type in ("symx", "sym")):
        problem.add_constraint(V=problem.get_displacement_function_space().sub(0), sub_domains=boundaries_mf, sub_domain_id=xmin_id, val=0.)
    if (const_type in ("symy", "sym")) and (dim >= 2):
        problem.add_constraint(V=problem.get_displacement_function_space().sub(1), sub_domains=boundaries_mf, sub_domain_id=ymin_id, val=0.)
    if (const_type in ("symz", "sym")) and (dim >= 3):
        problem.add_constraint(V=problem.get_displacement_function_space().sub(2), sub_domains=boundaries_mf, sub_domain_id=zmin_id, val=0.)
    if (const_type in ("blox")):
        problem.add_constraint(V=problem.get_displacement_function_space(), sub_domains=boundaries_mf, sub_domain_id=xmin_id, val=[0.]*dim)
    if (const_type in ("bloy")):
        problem.add_constraint(V=problem.get_displacement_function_space(), sub_domains=boundaries_mf, sub_domain_id=ymin_id, val=[0.]*dim)
    if (const_type in ("bloz")):
        problem.add_constraint(V=problem.get_displacement_function_space(), sub_domains=boundaries_mf, sub_domain_id=zmin_id, val=[0.]*dim)

    n_steps = step_params.get("n_steps", 1)
    Deltat_lst = step_params.get("Deltat_lst", [step_params.get("Deltat", 1.)/n_steps]*n_steps)
    dt_ini_lst = step_params.get("dt_ini_lst", [step_params.get("dt_ini", 1.)/n_steps]*n_steps)
    dt_min_lst = step_params.get("dt_min_lst", [step_params.get("dt_min", 1.)/n_steps]*n_steps)

    load_type = load_params.get("type", "disp")

    if   (load_type == "disp"):
        u_lst = load_params.get("u_lst", [(k_step+1)*load_params.get("u", +0.5)/n_steps for k_step in range(n_steps)])
    elif (load_type in ("volu0", "volu")):
        f_lst = load_params.get("f_lst", [(k_step+1)*load_params.get("f", +0.5)/n_steps for k_step in range(n_steps)])
    elif (load_type in ("surf0", "surf")):
        f_lst = load_params.get("f_lst", [(k_step+1)*load_params.get("f", +1.0)/n_steps for k_step in range(n_steps)])
    elif (load_type in ("pres0", "pres0_multi", "pres0_inertia", "pres")):
        p_lst = load_params.get("f_lst", [(k_step+1)*load_params.get("f", -0.5)/n_steps for k_step in range(n_steps)])
    elif (load_type in ("pgra0", "pgra")):
        X0_lst = load_params.get("X0_lst", [load_params.get("X0",      [0.5]* dim   )]*n_steps)
        N0_lst = load_params.get("N0_lst", [load_params.get("X0", [1.]+[0.0]*(dim-1))]*n_steps)
        P0_lst = load_params.get("P0_lst", [(k_step+1)*load_params.get("P0", -0.50)/n_steps for k_step in range(n_steps)])
        DP_lst = load_params.get("DP_lst", [(k_step+1)*load_params.get("P0", -0.25)/n_steps for k_step in range(n_steps)])
    elif (load_type == "tens"):
        gamma_lst = load_params.get("gamma_lst", [(k_step+1)*load_params.get("gamma", 0.01)/n_steps for k_step in range(n_steps)])
            
    for k_step in range(n_steps):

        Deltat = Deltat_lst[k_step]
        dt_ini = dt_ini_lst[k_step]
        dt_min = dt_min_lst[k_step]

        k_step = problem.add_step(
            Deltat=Deltat,
            dt_ini=dt_ini,
            dt_min=dt_min)

        if (load_type == "disp"):
            u = u_lst[k_step]
            u_old = u_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_constraint(
                V=problem.get_displacement_function_space().sub(0),
                sub_domains=boundaries_mf,
                sub_domain_id=xmax_id,
                val_ini=u_old, val_fin=u,
                k_step=k_step)
        elif (load_type == "volu0"):
            f = f_lst[k_step]
            f_old = f_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_volume_force0_loading_operator(
                measure=problem.dV,
                F_ini=[f_old]+[0.]*(dim-1), F_fin=[f]+[0.]*(dim-1),
                k_step=k_step)
        elif (load_type == "volu"):
            f = f_lst[k_step]
            f_old = f_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_volume_force_loading_operator(
                measure=problem.dV,
                F_ini=[f_old]+[0.]*(dim-1), F_fin=[f]+[0.]*(dim-1),
                k_step=k_step)
        elif (load_type == "surf0"):
            f = f_lst[k_step]
            f_old = f_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_force0_loading_operator(
                measure=problem.dS(xmax_id),
                F_ini=[f_old]+[0.]*(dim-1), F_fin=[f]+[0.]*(dim-1),
                k_step=k_step)
        elif (load_type == "surf"):
            f = f_lst[k_step]
            f_old = f_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_force_loading_operator(
                measure=problem.dS(xmax_id),
                F_ini=[f_old]+[0.]*(dim-1), F_fin=[f]+[0.]*(dim-1),
                k_step=k_step)
        elif (load_type == "pres0"):
            p = p_lst[k_step]
            p_old = p_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(xmax_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
        elif (load_type == "pres0_multi"):
            p = p_lst[k_step]
            p_old = p_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(xmax_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
            problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(ymax_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
            if (dim==3): problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(zmax_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
        elif (load_type == "pres0_inertia"):
            p = p_lst[k_step]
            p_old = p_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_inertia_operator(
                measure=problem.dV,
                rho_val=1e-2,
                k_step=k_step)
            problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(xmin_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
            problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(xmax_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
            problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(ymin_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
            problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(ymax_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
            if (dim==3): problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(zmin_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
            if (dim==3): problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(zmax_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
        elif (load_type == "pres"):
            p = p_lst[k_step]
            p_old = p_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_pressure_loading_operator(
                measure=problem.dS(xmax_id),
                P_ini=p_old, P_fin=p,
                k_step=k_step)
        elif (load_type == "pgra0"):
            X0 = X0_lst[k_step]
            N0 = N0_lst[k_step]
            P0 = P0_lst[k_step]
            DP = DP_lst[k_step]
            P0_old = P0_lst[k_step-1] if (k_step > 0) else 0.
            DP_old = DP_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_pressure_gradient0_loading_operator(
                measure=problem.dS(),
                X0_val=X0,
                N0_val=N0,
                P0_ini=P0_old, P0_fin=P0,
                DP_ini=DP_old, DP_fin=DP,
                k_step=k_step)
        elif (load_type == "pgra"):
            X0 = X0_lst[k_step]
            N0 = N0_lst[k_step]
            P0 = P0_lst[k_step]
            DP = DP_lst[k_step]
            P0_old = P0_lst[k_step-1] if (k_step > 0) else 0.
            DP_old = DP_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_pressure_gradient_loading_operator(
                measure=problem.dS(),
                X0_val=X0,
                N0_val=N0,
                P0_ini=P0_old, P0_fin=P0,
                DP_ini=DP_old, DP_fin=DP,
                k_step=k_step)
        elif (load_type == "tens"):
            gamma = gamma_lst[k_step]
            gamma_old = gamma_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_tension_loading_operator(
                measure=problem.dS,
                gamma_ini=gamma_old, gamma_fin=gamma,
                k_step=k_step)

    ################################################# Quantities of Interest ###

    problem.add_global_strain_qois()
    problem.add_global_stress_qois()
    if (incomp): problem.add_global_pressure_qoi()
    if (inverse==0) and (dim==2): problem.add_global_out_of_plane_stress_qois()

    ################################################################# Solver ###

    solver = dmech.NonlinearSolver(
        problem=problem,
        parameters={
            "sol_tol":[1e-6]*len(problem.subsols),
            "n_iter_max":32},
        relax_type="constant",
        write_iter=0)

    integrator = dmech.TimeIntegrator(
        problem=problem,
        solver=solver,
        parameters={
            "n_iter_for_accel":4,
            "n_iter_for_decel":16,
            "accel_coeff":2,
            "decel_coeff":2},
        print_out=res_basename*verbose,
        print_sta=res_basename*verbose,
        write_qois=res_basename+"-qois",
        write_qois_limited_precision=1,
        write_sol=res_basename*verbose,
        write_vtus=res_basename*verbose,
        write_vtus_with_preserved_connectivity=write_vtus_with_preserved_connectivity)

    success = integrator.integrate()
    assert (success),\
        "Integration failed. Aborting."

    integrator.close()

    if (get_results):
        return (problem.get_displacement_subsol().func, dolfin.Measure("dx", domain=mesh))



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Felipe Álvarez-Barrientos, 2019-2021                                 ###
###                                                                          ###
### Pontificia Universidad Católica, Santiago, Chile                         ###
###                                                                          ###
###                                                                          ###
### And Mahdi Manoochehrtayebi, 2020-2024                                    ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### Inspired from https://comet-fenics.readthedocs.io/en/latest/             ###
###                        demo/periodic_homog_elas/periodic_homog_elas.html ###
###                                                                          ###
################################################################################

import dolfin
import numpy as np

################################################################################

class PeriodicSubDomain(dolfin.SubDomain):



    def __init__(self,
            dim,
            bbox,
            vertices=None,
            tol=None):

        self.dim = dim

        self.xmin = bbox[0]
        self.xmax = bbox[1]
        if (self.dim >= 2):
            self.ymin = bbox[2]
            self.ymax = bbox[3]
        if (self.dim >= 3):
            self.zmin = bbox[4]
            self.zmax = bbox[5]

        if (tol is None):
            self.tol = (self.xmax - self.xmin)**2
            if (self.dim >= 2):
                self.tol += (self.ymax - self.ymin)**2
            if (self.dim >= 3):
                self.tol += (self.zmax - self.zmin)**2
            self.tol = (self.tol)**(1/2)
            self.tol *= 1e-2
        else:
            self.tol = tol
       
        dolfin.SubDomain.__init__(self, map_tol=self.tol)

        if (self.dim == 2):
            self.inside = self.inside_2D
            self.map = self.map_2D
        elif (self.dim == 3):
            self.inside = self.inside_3D
            self.map = self.map_3D
        else:
            assert (0),\
                "dim must be 2 or 3. Aborting."

    
        if (vertices is None):
            vertices = np.array([[self.xmin, self.ymin],
                                 [self.xmax, self.ymin],
                                 [self.xmax, self.ymax],
                                 [self.xmin, self.ymax]])
        # if (vertices is not None):
        self.vv = vertices
        self.a1 = self.vv[1,:]-self.vv[0,:] # first vector generating periodicity
        self.a2 = self.vv[3,:]-self.vv[0,:] # second vector generating periodicity
        # check if UC vertices form indeed a parallelogram
        assert np.linalg.norm(self.vv[2,:] - self.vv[3,:] - self.a1) <= self.tol
        assert np.linalg.norm(self.vv[2,:] - self.vv[1,:] - self.a2) <= self.tol



    # def inside_2D(self, x, on_boundary):
    #     return bool(on_boundary and (dolfin.near(x[0], self.xmin, self.tol) or dolfin.near(x[1], self.ymin, self.tol)) and not (dolfin.near(x[0], self.xmax, self.tol) or dolfin.near(x[1], self.ymax, self.tol)))



    def inside_2D(self, x, on_boundary):
        # return True if on left or bottom boundary AND NOT on one of the
        # bottom-right or top-left vertices
        return bool((dolfin.near(x[0], self.vv[0,0] + x[1]*self.a2[0]/self.vv[3,1], self.tol) or
                    dolfin.near(x[1], self.vv[0,1] + x[0]*self.a1[1]/self.vv[1,0], self.tol)) and
                    (not ((dolfin.near(x[0], self.vv[1,0], self.tol) and dolfin.near(x[1], self.vv[1,1], self.tol)) or
                    (dolfin.near(x[0], self.vv[3,0], self.tol) and dolfin.near(x[1], self.vv[3,1], self.tol)))) and on_boundary)



    def inside_3D(self, x, on_boundary):
        return bool(on_boundary and (dolfin.near(x[0], self.xmin, self.tol) or dolfin.near(x[1], self.ymin, self.tol) or dolfin.near(x[2], self.zmin, self.tol)) and not (dolfin.near(x[0], self.xmax, self.tol) or dolfin.near(x[1], self.ymax, self.tol) or dolfin.near(x[2], self.zmax, self.tol)))



    # def map_2D(self, x, y):
    #     # Vertex
    #     if  (dolfin.near(x[0], self.xmax, self.tol) \
    #     and  dolfin.near(x[1], self.ymax, self.tol))\
    #     or  (dolfin.near(x[0], self.xmax, self.tol) \
    #     and  dolfin.near(x[1], self.ymin, self.tol))\
    #     or  (dolfin.near(x[0], self.xmin, self.tol) \
    #     and  dolfin.near(x[1], self.ymax, self.tol)):
    #         y[0] = self.xmin
    #         y[1] = self.ymin
    #     # Edges
    #     elif dolfin.near(x[0], self.xmax, self.tol):
    #         y[0] = self.xmin
    #         y[1] = x[1]
    #     elif dolfin.near(x[1], self.ymax, self.tol):
    #         y[0] = x[0]
    #         y[1] = self.ymin
    #     # FAB: Else clause is there to have the point always end up mapped somewhere, cf. https://fenicsproject.discourse.group/t/periodic-boundary-class-for-2-target-domains/99/2
    #     else:
    #         y[0] = -1.
    #         y[1] = -1.



    def map_2D(self, x, y):
        if dolfin.near(x[0], self.vv[2,0], self.tol) and dolfin.near(x[1], self.vv[2,1], self.tol): # if on top-right corner
            y[0] = x[0] - (self.a1[0]+self.a2[0])
            y[1] = x[1] - (self.a1[1]+self.a2[1])
        elif dolfin.near(x[0], self.vv[1,0] + x[1]*self.a2[0]/self.vv[2,1], self.tol): # if on right boundary
            y[0] = x[0] - self.a1[0]
            y[1] = x[1] - self.a1[1]
        else:   # should be on top boundary
            y[0] = x[0] - self.a2[0]
            y[1] = x[1] - self.a2[1]



    def map_3D(self, x, y):
        # Vertex
        if  (dolfin.near(x[0], self.xmax, self.tol) \
        and  dolfin.near(x[1], self.ymax, self.tol) \
        and  dolfin.near(x[2], self.zmax, self.tol))\
        or  (dolfin.near(x[0], self.xmax, self.tol) \
        and  dolfin.near(x[1], self.ymax, self.tol) \
        and  dolfin.near(x[2], self.zmin, self.tol))\
        or  (dolfin.near(x[0], self.xmax, self.tol) \
        and  dolfin.near(x[1], self.ymin, self.tol) \
        and  dolfin.near(x[2], self.zmax, self.tol))\
        or  (dolfin.near(x[0], self.xmax, self.tol) \
        and  dolfin.near(x[1], self.ymin, self.tol) \
        and  dolfin.near(x[2], self.zmin, self.tol))\
        or  (dolfin.near(x[0], self.xmin, self.tol) \
        and  dolfin.near(x[1], self.ymax, self.tol) \
        and  dolfin.near(x[2], self.zmax, self.tol))\
        or  (dolfin.near(x[0], self.xmin, self.tol) \
        and  dolfin.near(x[1], self.ymax, self.tol) \
        and  dolfin.near(x[2], self.zmin, self.tol))\
        or  (dolfin.near(x[0], self.xmin, self.tol) \
        and  dolfin.near(x[1], self.ymin, self.tol) \
        and  dolfin.near(x[2], self.zmax, self.tol)):
            y[0] = self.xmin
            y[1] = self.ymin
            y[2] = self.zmin
        # Edges
        elif (dolfin.near(x[0], self.xmax, self.tol) \
        and   dolfin.near(x[1], self.ymax, self.tol))\
        or   (dolfin.near(x[0], self.xmax, self.tol) \
        and   dolfin.near(x[1], self.ymin, self.tol))\
        or   (dolfin.near(x[0], self.xmin, self.tol) \
        and   dolfin.near(x[1], self.ymax, self.tol)):
            y[0] = self.xmin
            y[1] = self.ymin
            y[2] = x[2]
        elif (dolfin.near(x[0], self.xmax, self.tol) \
        and   dolfin.near(x[2], self.zmax, self.tol))\
        or   (dolfin.near(x[0], self.xmax, self.tol) \
        and   dolfin.near(x[2], self.zmin, self.tol))\
        or   (dolfin.near(x[0], self.xmin, self.tol) \
        and   dolfin.near(x[2], self.zmax, self.tol)):
            y[0] = self.xmin
            y[1] = x[1]
            y[2] = self.zmin
        elif (dolfin.near(x[1], self.ymax, self.tol) \
        and   dolfin.near(x[2], self.zmax, self.tol))\
        or   (dolfin.near(x[1], self.ymax, self.tol) \
        and   dolfin.near(x[2], self.zmin, self.tol))\
        or   (dolfin.near(x[1], self.ymin, self.tol) \
        and   dolfin.near(x[2], self.zmax, self.tol)):
            y[0] = x[0]
            y[1] = self.ymin
            y[2] = self.zmin
        # Faces
        elif dolfin.near(x[0], self.xmax, self.tol):
            y[0] = self.xmin
            y[1] = x[1]
            y[2] = x[2]
        elif dolfin.near(x[1], self.ymax, self.tol):
            y[0] = x[0]
            y[1] = self.ymin
            y[2] = x[2]
        elif dolfin.near(x[2], self.zmax, self.tol):
            y[0] = x[0]
            y[1] = x[1]
            y[2] = self.zmin
        # FAB: Else clause is there to have the point always end up mapped somewhere, cf. https://fenicsproject.discourse.group/t/periodic-boundary-class-for-2-target-domains/99/2
        else:
            y[0] = -1000.
            y[1] = -1000.
            y[2] = -1000.



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech

################################################################################

class QOI():



    def __init__(self,
            name,
            expr,
            norm=1.,
            constant=0.,
            divide_by_dt=False,
            form_compiler_parameters={},
            point=None,
            update_type="assembly"):

        self.name                     = name
        self.expr                     = expr
        self.norm                     = norm
        self.constant                 = constant
        self.divide_by_dt             = divide_by_dt
        self.form_compiler_parameters = form_compiler_parameters
        self.point                    = point

        if (update_type == "assembly"):
            self.update = self.update_assembly
        elif (update_type == "direct"):
            self.update = self.update_direct



    def update_assembly(self, dt=None):

        # print(self.name)
        # print(self.expr)
        # print(self.form_compiler_parameters)

        self.value = dolfin.assemble(
            self.expr,
            form_compiler_parameters=self.form_compiler_parameters)

        self.value += self.constant
        self.value /= self.norm

        if (self.divide_by_dt):
            assert (dt != 0),\
                "dt (="+str(dt)+") should be non zero. Aborting."
            self.value /= dt



    def update_direct(self, dt=None):
        
        self.value = self.expr(self.point)

        self.value += self.constant
        self.value /= self.norm

        if (self.divide_by_dt) and (dt is not None):
            self.value /= dt



#coding=utf8

################################################################################
###                                                                          ###
### Created by Mahdi Manoochehrtayebi, 2020-2024                             ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Martin Genet, 2018-2024                                              ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class DeformedFluidVolumeOperator(Operator):

    def __init__(self,
            vf,
            vf_test,
            kinematics,
            N,
            dS,
            U_tot,
            X, 
            measure):

        self.kinematics = kinematics
        self.U_tot = U_tot
        self.X = X
        self.N = N
        self.dS = dS
        self.measure = measure

        PN = self.kinematics.J * dolfin.dot(dolfin.inv(self.kinematics.F.T), self.N)
        
        self.res_form = ((vf + dolfin.assemble(dolfin.inner(self.U_tot + self.X, PN) * self.dS(0))/2) * vf_test) * self.measure # MG20230203: This is not correct. Need to check.



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class NeoHookeanMooneyRivlinElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters,
            decoup=False):

        self.kinematics = kinematics

        C1,C2 = self.get_C1_and_C2_from_parameters(parameters) # MG20220318: This is different from computing C1 & C2 separately…
        parameters["C1"] = C1
        parameters["C2"] = C2

        self.nh = dmech.NeoHookeanElasticMaterial(kinematics, parameters, decoup)
        self.mr = dmech.MooneyRivlinElasticMaterial(kinematics, parameters, decoup)

        self.Psi   = self.nh.Psi   + self.mr.Psi
        self.Sigma = self.nh.Sigma + self.mr.Sigma
        if (self.kinematics.dim == 2):
            self.Sigma_ZZ = self.nh.Sigma_ZZ + self.mr.Sigma_ZZ
        self.P     = self.nh.P     + self.mr.P
        self.sigma = self.nh.sigma + self.mr.sigma



    # def get_free_energy(self, *args, **kwargs):

    #     Psi_nh, Sigma_nh = self.nh.get_free_energy(*args, **kwargs)
    #     Psi_mr, Sigma_mr = self.mr.get_free_energy(*args, **kwargs)

    #     Psi   = Psi_nh   + Psi_mr
    #     Sigma = Sigma_nh + Sigma_mr

    #     return Psi, Sigma
